"""
Nasse

© Anime no Sekai — 2023
"""
import logging
import os
import pathlib
import sys
import threading
import typing
import urllib.parse
import dataclasses

import flask
import rich.progress
import rich.status
import watchdog.events
import watchdog.observers

from nasse import config, docs, models, receive, request, utils
from nasse.config import NasseConfig
from nasse.localization.base import Localization
from nasse.response import exception_to_response
from nasse.servers import ServerBackend
from nasse.servers.flask import Flask


class FileEventHandler(watchdog.events.FileSystemEventHandler):
    """An internal file event handler for the debug mode"""

    def __init__(self, callback: typing.Callable, watch: typing.List[pathlib.Path], ignore: typing.List[pathlib.Path], config: NasseConfig = None) -> None:
        super().__init__()
        self.config = config or NasseConfig()
        self.callback = callback
        self.watch = [str(file) for file in watch]
        self.ignore = [str(file) for file in ignore]

    def on_modified(self, event):
        src_path = str(pathlib.Path(str(event.src_path)).resolve())
        if src_path not in self.watch or src_path in self.ignore:
            return
        self.config.logger.log("{path} modified".format(path=src_path))
        self.callback()


class Nasse:
    """The Nasse web server object"""

    def __init__(self,
                 name: typing.Optional[str] = None,
                 config: typing.Optional[NasseConfig] = None,
                 flask_options: typing.Optional[dict] = None,
                 *args, **kwargs) -> None:
        """
        # Nasse

        Examples
        ---------
        >>> from nasse import Nasse
        >>> app = Nasse()

        ...is the most minimalist usage, but we obviously recommend customizing the parameters:

        >>> app = Nasse(name="New Server", cors=["https://new-server.com", "https://beta-new-server.com"])

        Parameters
        -----------
        name: str, default = None
            Defines the name of the server
        config: NasseConfig, default = None
            A NasseConfig object, with all of the configuration for the server
        flask_options: dict, default = None
            A dictionary with the options to pass to Flask when creating the Flask instance
        *args
            Arguments to give to NasseConfig
        **kwargs
            Keyword arguments to give to NasseConfig
        """
        if isinstance(name, NasseConfig):
            config = name
            name = None
        if config:
            config_kwargs = config.__dict__
            config_kwargs.update(kwargs)
            config_kwargs["name"] = name or config_kwargs.get("app", "Nasse")
            # config_kwargs.pop("VERSION", None)
            self.config = NasseConfig(*args, **config_kwargs)
        else:
            self.config = NasseConfig(name=name or "Nasse", *args, **kwargs)

        self.config.logger.config = self.config
        utils.logging.logger = self.config.logger

        if self.config.debug:
            threading.settrace(utils.logging._generate_trace(self.config))

        if isinstance(self.config.account_management, type):
            self.config.account_management = self.config.account_management()

        self.flask = flask.Flask(self.config.name, **(flask_options or {}))

        self.endpoints = {}

        # security
        self.flask.config["MAX_CONTENT_LENGTH"] = int(self.config.max_request_size) if self.config.max_request_size is not None else None

        # events
        self.flask.before_request(lambda: self.before_request())
        # the lambda function, acting as a proxy between the real after_request function is here to let the user change the function as wanted
        self.flask.after_request(lambda response: self.after_request(response))

        self.flask.register_error_handler(Exception, self.handle_exception)

        if self.config.compress:
            import flask_compress
            flask_compress.Compress(self.flask)

        logging.getLogger('werkzeug').disabled = True
        self.flask.logger.disabled = True

        # on debug
        self._observer = None

    def __repr__(self) -> str:
        return "Nasse({name})".format(name=self.config.name)

    @property
    def logger(self) -> utils.logging.Logger:
        return self.config.logger

    def log(self, *msg, **kwargs):
        return self.logger.log(*msg, **kwargs)

    def route(self,
              path: typing.Optional[typing.Union[typing.Callable, str]] = None,

              name: str = "",
              category: str = "",
              sub_category: str = "",
              description: models.Types.MethodVariant[str] = None,
              base_dir: typing.Union[pathlib.Path, str, None] = None,
              endpoint: typing.Optional[typing.Union[models.Endpoint, typing.Mapping]] = None,

              # Request,
              methods: models.Types.OptionalIterable[models.Types.Method.Any] = "*",
              login: models.Types.MethodVariant[models.Types.OptionalIterable[models.Login]] = None,

              # User Sent,
              parameters: models.Types.MethodVariant[models.Types.OptionalIterable[models.Parameter]] = None,
              headers: models.Types.MethodVariant[models.Types.OptionalIterable[models.Header]] = None,
              cookies: models.Types.MethodVariant[models.Types.OptionalIterable[models.Cookie]] = None,
              dynamics: models.Types.MethodVariant[models.Types.OptionalIterable[models.Dynamic]] = None,

              # Response,
              json: bool = True,
              returns: models.Types.MethodVariant[models.Types.OptionalIterable[models.Return]] = None,
              errors: models.Types.MethodVariant[models.Types.OptionalIterable[models.Error]] = None,
              flask_options: typing.Optional[dict] = None) -> typing.Callable[..., models.Endpoint]:
        """
        Use this function to declare new endpoints

        Examples
        --------
        >>> # Can be used like so:
        >>> from nasse import Nasse, Param
        >>> app = Nasse()
        >>> 
        >>> @app.route(path="/hello", params=Param(name="username", description="the person welcomed"))
        ... def hello(params: dict):
        ...     if "username" in params:
        ...         return f"Hello {params['username']}"
        ...     return "Hello world"

        Parameters
        -----------
        flask_options: dict
            If needed, extra options to give to flask.Flask
        """
        flask_options = dict(flask_options or {})

        def decorator(handler):
            # if `path` callable, we are in an argument-less decoration
            new_endpoint = models.Endpoint(handler=handler,
                                           name=name,
                                           category=category,
                                           sub_category=sub_category,
                                           description=description,
                                           base_dir=base_dir,
                                           endpoint=endpoint or (path if isinstance(path, models.Endpoint) else None),
                                           path=path if not callable(path) and not isinstance(path, models.Endpoint) else None,
                                           methods=methods,
                                           login=login,
                                           parameters=parameters,
                                           headers=headers,
                                           cookies=cookies,
                                           dynamics=dynamics,
                                           json=json,
                                           returns=returns,
                                           errors=errors)

            try:
                flask_options["methods"] = (new_endpoint.methods
                                            if "*" not in new_endpoint.methods
                                            else typing.get_args(models.Types.Method.Standard))
            except Exception:
                pass

            self.flask.add_url_rule(new_endpoint.path,
                                    flask_options.pop("endpoint", None),
                                    receive.Receive(self, new_endpoint),
                                    **flask_options)

            self.endpoints[new_endpoint.path] = new_endpoint
            return handler

        if callable(path):
            result = decorator(path)  # called without arguments
            return result

        return decorator  # called with arguments

    def run(self,
            host: typing.Optional[str] = None,
            port: typing.Optional[int] = None,
            server: typing.Type[ServerBackend] = Flask,
            watch: typing.Optional[typing.List[str]] = None,
            ignore: typing.Optional[typing.List[str]] = None,
            status: bool = True,
            *args, **kwargs) -> None:
        """
        Runs the application by binding to an address and answering to clients.

        Note: If a `debug` argument is passed to the underlying server backend using `kwargs`, it will be used for `config.debug`

        Parameters
        ----------
        host: str, optional
            The host to run the server on
        port: int, optional
            The port to make the server listen at
        server: Type[ServerBackend], default=Flask
            The server backend to use
        watch: list[str], default=["**/*.py"]
            A list of glob expressions of files to watch for changes in debug mode
        ignore: list[str], optional
            A list of glob expressions of files to ignore changes for in debug mode
        status: bool, default=True
            Whether to show the progress status (time since the server first ran, URL listened, etc.).
            This might come handy in complex environments where you are already using
            a `rich.progress.Progress object, which might conflict.
        """
        if watch is None:
            watch = ["**/*.py"]

        class MockProgress:
            """Replaces the logger"""

            def __overwrite__(self, *args, **kwargs):
                return self

            add_task = __overwrite__
            update = __overwrite__
            __enter__ = __overwrite__
            __exit__ = __overwrite__

        with (rich.progress.Progress(transient=True) if status else MockProgress()) as progress:
            main_task = progress.add_task("Setting up the environment", total=None)

            if host is not None:
                self.config.host = host
            if port is not None:
                self.config.port = int(port)
            if "debug" in kwargs:
                self.config.debug = bool(kwargs["debug"])

            if self.config.debug:
                self.config.logger.warn("DEBUG MODE IS ENABLED")
                # Configuring auto-restart
                try:
                    watching = []
                    ignoring = []
                    for storage, data in [(watching, watch), (ignoring, ignore or [])]:
                        for file in data:
                            file = str(file)
                            path = pathlib.Path(file)
                            if path.is_file():
                                storage.append(path.resolve())
                            elif path.is_dir():
                                storage.extend(child.resolve() for child in path.iterdir())
                            else:
                                storage.extend(child.resolve() for child in pathlib.Path().glob(file))
                    self._observer = watchdog.observers.Observer()
                    self._observer.schedule(FileEventHandler(callback=self.restart, watch=watching,
                                            ignore=ignoring, config=self.config), ".", recursive=True)
                    self._observer.start()
                except Exception as err:
                    self.config.logger.warn(f"Couldn't set up the file changes watcher ({err.__class__.__name__}: {err})")
                # Configuring debug endpoints
                try:
                    def endpoints():
                        """Returns back all of the defined endpoints"""
                        results = {"endpoints": []}
                        for endpoint in self.endpoints.values():
                            # endpoint: models.Endpoint
                            result = {
                                "handler": endpoint.handler.__name__,
                                "name": endpoint.name,
                                "category": endpoint.category,
                                "sub_category": endpoint.sub_category,
                                "description": endpoint.description,
                                "base_dir": endpoint.base_dir,
                                "path": endpoint.path,
                                "methods": endpoint.methods,
                                "json": endpoint.json
                            }
                            for element in ("login", "parameters", "headers", "cookies", "dynamics", "returns", "errors"):
                                result[element] = {key: [dataclasses.asdict(val) for val in value]
                                                   for key, value in getattr(endpoint, element).items()}

                            results["endpoints"].append(result)
                        return 200, results

                    self.route("/@nasse/endpoints",
                               name="Endpoints",
                               category="@Nasse Debug")(endpoints)
                except Exception as err:
                    if self.config.logger:
                        self.config.logger.warn(f"Couldn't set up the debug endpoints ({err.__class__.__name__}: {err})")

            self.instance = server(app=self, config=self.config)

        # Main Loop
        with (rich.progress.Progress(*(rich.progress.TextColumn("[progress.description]{task.description}"),
                                       rich.progress.TextColumn("—"),
                                       rich.progress.TimeElapsedColumn()),
                                     transient=True) if status else MockProgress()) as progress:
            progress.add_task(description=f'🍡 {self.config.name} is running on http://{self.config.host}:{self.config.port}')
            self.config.logger.log("🎏 Press {cyan}Ctrl+C{normal} to quit")
            self.config.logger.log(f"🌍 Binding to {{magenta}}{self.config.host}:{self.config.port}{{normal}}")
            # spinner = rich.progress.SpinnerColumn(spinner_name="simpleDotsScrolling", style="gray")
            # spinner.spinner.frames = ["・　　", "・・　", "・・・", "　・・", "　　・", "　　　"]
            self.instance.run(*args, **kwargs)

    def restart(self):
        """Restarts the current python process"""
        self.config.logger.log("Restarting... 🎐")
        if self._observer:
            self.config.logger.debug("Waiting for watchdog to terminate")
            try:
                self._observer.stop()
                self._observer.join()
            except Exception:
                pass
        self.config.logger.log("Stopping the server instance")
        self.instance.stop()
        os.execv(sys.executable, [sys.executable] + sys.argv)

    def handle_exception(self, e):
        """
        Handles exception for flask.Flask
        """
        try:
            try:
                message, error, code = exception_to_response(e)
            except Exception:
                message, error, code = "An error occured on the server", "SERVER_ERROR", 500
            result = {"success": False, "message": message, "error": error, "data": {}}
            content_type = "application/json"
            try:
                if utils.sanitize.remove_spaces(flask.request.values.get("format", "json")).lower() in {"xml", "html"}:
                    body = utils.xml.encode(data=result, minify=True)
                    content_type = "application/xml"
                raise ValueError("Not XML")
            except Exception:
                body = utils.json.minified_encoder.encode(result)
            return flask.Response(response=body, status=code, content_type=content_type)
        except Exception:
            return flask.Response(response='{"success": false, "message": "An error occured on the server", "error": "SERVER_ERROR", "data": {}', status=500, content_type="application/json")

    def before_request(self):
        """
        Internal function called when before receiving a response

        Might be useful to check things (like a database) before the request

        You won't need this function to perform validation if you are consistently
        and rightfully documenting your endpoints as Nasse will take care of this for you.
        """
        return

    def after_request(self, response: flask.Response):
        """
        Internal function called before sending back a response
        It applies multiple security headers to ensure HTTPS, CORS, etc.

        Parameters
        -----------
        response: flask.flask.Response
            The response to send back

        Returns
        --------
        flask.Response:
            The response to send
        """
        try:
            # Ensuring HTTPS (for a year)
            response.headers["Strict-Transport-Security"] = "max-age=31536000; includeSubDomains; preload"

            # Managing CORS
            # Allowing the right methods
            try:
                if flask.request.method.upper() == "OPTIONS":
                    current_endpoint = self.endpoints.get(flask.request.url_rule.rule,
                                                          None)
                    if current_endpoint is not None:
                        try:
                            response.headers["Access-Control-Allow-Methods"] = ",".join(current_endpoint.methods)
                        except Exception:
                            from traceback import print_exc
                            print_exc()
                            self.config.logger.warn("An error occured while setting the Access-Control-Allow-Methods header")
                        try:
                            requested_headers = [header.lower() for header in utils.sanitize.remove_spaces(
                                flask.request.headers.get("Access-Control-Request-Headers", "")).split(",")]
                            endpoint_headers = [header.name.lower() for header in current_endpoint.headers]

                            # login_rules = current_endpoint.login.get(flask.request.method.upper(), current_endpoint.login.get("*", None))
                            # if login_rules is not None and not login_rules.no_login:
                            #     endpoint_headers.append("authorization")
                            endpoint_headers.append("authorization")

                            response.headers["Access-Control-Allow-Headers"] = ",".join(
                                (header for header in requested_headers if header in endpoint_headers))
                        except Exception:
                            from traceback import print_exc
                            print_exc()
                            self.config.logger.warn("An error occured while setting the Access-Control-Allow-Headers header")
                    else:
                        self.config.logger.warn("We couldn't verify the current endpoint informations on an OPTIONS request")
                    response.headers["Access-Control-Max-Age"] = 86400
            except Exception:
                self.config.logger.warn("An error occured while setting some CORS headers")

            # Allowing the right origins
            if self.config.cors:
                if "*" in self.config.cors:
                    origin = flask.request.environ.get("HTTP_ORIGIN", None)
                    if origin is not None:
                        response.headers["Vary"] = "Origin"
                        response.headers["Access-Control-Allow-Origin"] = origin
                    else:
                        response.headers["Access-Control-Allow-Origin"] = "*"
                else:
                    response.headers["Vary"] = "Origin"
                    request_origin = flask.request.environ.get(
                        "HTTP_ORIGIN", None)
                    if request_origin in self.config.cors:
                        response.headers["Access-Control-Allow-Origin"] = request_origin
                    else:
                        response.headers["Access-Control-Allow-Origin"] = self.config.cors[0]
            # Might need to allow the right headers
            # ...

            response.headers["Server"] = utils.formatter.format(self.config.server_header, config=self.config)

            # response.headers["Server"] = "Nasse/{version} ({name})".format(version=self.config.VERSION,
            #                                                                name=self.config.name)

        except Exception:
            # would be bad if the `after_request` function raises an exception, especially when used as the `teardown_request`
            pass

        return response

    def make_docs(self, base_dir: typing.Optional[typing.Union[pathlib.Path, str]] = None,
                  curl: bool = True, javascript: bool = True, python: bool = True,
                  localization: typing.Union[typing.Type[Localization], Localization] = Localization):
        """
        Creates the documentation for your API/Server

        Parameters
        ----------
        base_dir: str | Path
            The path where the docs will be outputted.
            This shouldn't be a directory where the `postman` docs and all of the sections will be outputted.
        curl: bool
            Whether or not to generate the curl examples
        javascript: bool
            Whether or not to generate the javascript examples
        python: bool
            Whether or not to generate the python examples
        localization: Localization
            The language for the docs
        """
        with rich.progress.Progress(rich.progress.SpinnerColumn(),
                                    *rich.progress.Progress.get_default_columns(),
                                    transient=True) as progress:
            main_task = progress.add_task("Creating the API Reference Documentation", total=5)
            self.config.logger.hide("Creating the API Reference Documentation")

            docs_path = pathlib.Path(base_dir or pathlib.Path() / "docs")
            docs_path.mkdir(parents=True, exist_ok=True)

            postman_path = docs_path / "Postman"
            postman_path.mkdir(parents=True, exist_ok=True)

            sections_path = docs_path / localization.sections
            sections_path.mkdir(parents=True, exist_ok=True)

            progress.advance(main_task)

            # Initializing the resulting string by prepending the header
            result = localization.getting_started_header.format(name=self.config.name, id=self.config.id)

            result += "## {localization__index}\n\n".format(localization__index=localization.index)

            # Sorting the sections alphabetically
            sections = sorted({endpoint.category for endpoint in self.endpoints.values()})

            # Getting the endpoints for each section
            sections_registry = {}
            for section in sections:
                for endpoint in self.endpoints.values():
                    if endpoint.category == section:
                        try:
                            sections_registry[section].append(endpoint)
                        except Exception:
                            sections_registry[section] = [endpoint]

            progress.advance(main_task)

            headers_registry = []

            for section in sections_registry:
                current_link = docs.header.header_link(section, headers_registry)
                result += "- [{section}](./{localization__sections}/{section_url}.md#{link})\n".format(section=section,
                                                                                                       localization__sections=urllib.parse.quote(
                                                                                                           localization.sections, safe=''),
                                                                                                       section_url=section.replace(" ", "%20"),
                                                                                                       link=current_link)

                result += "\n".join(["  - [{endpoint}](./{localization__sections}/{section}.md#{link})".format(
                    endpoint=endpoint.name,
                    localization__sections=urllib.parse.quote(localization.sections, safe=""),
                    section=section.replace(" ", "%20"),
                    link=docs.header.header_link(endpoint.name, headers_registry)) for endpoint in sections_registry[section]]
                )
                result += "\n"

            progress.advance(main_task)

            with open(docs_path / "{localization__getting_started}.md".format(localization__getting_started=localization.getting_started), "w", encoding="utf8") as out:
                out.write(result)

            progress.advance(main_task)

            # Dumping all of the docs and creating the Postman Data
            for section in sections_registry:
                result = localization.section_header.format(name=section)
                # result += '''\n## {section}\n'''.format(section=section)
                result += "\n".join([docs.markdown.make_docs(endpoint, curl=curl, javascript=javascript, python=python, localization=localization, base_dir=base_dir)
                                    for endpoint in sections_registry[section]])
                with open(sections_path / "{section}.md".format(section=section), "w", encoding="utf-8") as out:
                    out.write(result)

                result = docs.postman.create_postman_data(self, section, sections_registry[section], localization=localization)
                with open(postman_path / "{section}.postman_collection.utils.json".format(section=section), "w", encoding="utf-8") as out:
                    out.write(utils.json.minified_encoder.encode(result))

            progress.advance(main_task)



"""
utils/types

Defining types to use on your web journey, such as enums and more
"""

import re
import typing


class LimitedString(str):
    """
    A string with conditions

    Variables to set: UPPER, LOWER, SPACES, THROW, REGEX, DEFAULT, LIMIT
    """
    UPPER = False
    """If the string should be converted to all upper case"""
    LOWER = False
    """If the string should be converted to all lower case"""
    SPACES = True
    """If we should allow spaces in the string"""
    THROW = False
    """It we should raise an error when there is an error"""
    REGEX: typing.Optional[re.Pattern] = None
    """A Regex pattern we need to follow for this string"""
    DEFAULT = ""
    """A default value if the regex pattern is not followed"""
    LIMIT = 3000
    """A number of characters limit"""

    def __new__(self, value: str, **kw):
        value = str(value)
        if self.UPPER:
            value = value.upper()
        elif self.LOWER:
            value = value.lower()
        if not self.SPACES:
            value = value.replace(" ", "")
        if self.REGEX is not None:
            if not self.REGEX.match(value):
                if self.THROW:
                    raise ValueError(f"{value} is not valid")
                value = self.DEFAULT
        if len(value) > self.LIMIT:
            if self.THROW:
                raise ValueError(f"The given value exceeds the {self.LIMIT} characters limit")
            value = value[:self.LIMIT]
        return str.__new__(self, value, **kw)


class StringEnum(str):
    """
    A string enum, only accepting certain values

    Variables to set: ACCEPTED, DEFAULT, UPPER, LOWER
    """
    ACCEPTED = tuple()
    DEFAULT = ""
    UPPER = True
    LOWER = False

    def __new__(cls, value: str, **kw):
        value = str(value).replace(" ", "")
        if cls.UPPER:
            value = value.upper()
        elif cls.LOWER:
            value = value.lower()
        if value not in cls.ACCEPTED:
            value = cls.DEFAULT
        return str.__new__(cls, value, **kw)


class Color(LimitedString):
    """Represents a hexadecimal color"""
    LIMIT = 7
    # REGEX = re.compile(r"^#[0-9a-fA-F]{6}$")
    REGEX = re.compile(r"^#(?:[0-9a-fA-F]{3}){1,2}$")


class Email(LimitedString):
    """Represents an email address"""
    REGEX: re.Pattern = re.compile(r"""(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])""")
    """
    Author
    ------
    StackOverflow Community
        https://stackoverflow.com/a/201378/11557354
    """


class HTTPMethod(StringEnum):
    """Represents an HTTP standard method"""
    ACCEPTED = ("GET", "HEAD", "POST",
                "PUT", "DELETE", "CONNECT",
                "OPTIONS", "TRACE", "PATCH")


class IPv4(LimitedString):
    """Represents an IPV4 address"""
    REGEX = re.compile(r"""^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)\.?\b){4}$""")
    """
    Author
    ------
    Danail Gabenski
        https://stackoverflow.com/a/36760050/11557354
    """


class IPv6(LimitedString):
    """Represents an IPV6 address"""
    REGEX = re.compile(r"""(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))""")
    """
    Author
    ------
    David M. Syzdek
        https://stackoverflow.com/a/17871737/11557354
    """


class PhoneNumber(LimitedString):
    """
    Represents a phone number
    """
    REGEX = re.compile(
        r"""\+(9[976]\d|8[987530]\d|6[987]\d|5[90]\d|42\d|3[875]\d|2[98654321]\d|9[8543210]|8[6421]|6[6543210]|5[87654321]|4[987654310]|3[9643210]|2[70]|7|1)\d{1,14}$""")
    """
    Author
    ------
    Eric
        https://stackoverflow.com/a/6967885/11557354
    """


class Filepath(LimitedString):
    """Represents a filepath"""
    REGEX = re.compile(r"""^(.+)\/([^\/]+)$""")
    """
    Author
    ------
    Paige Ruten
        https://stackoverflow.com/a/169021/11557354
    """


class URL(LimitedString):
    """Represents an URL"""
    REGEX = re.compile(r"""[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)""")
    """
    Author
    ------
    Daveo
        https://stackoverflow.com/a/3809435/11557354
    """



"""Redefines the base app"""
import json
import typing

from textual.app import App as TextualApp
from textual.binding import Binding, _Bindings

Bindings = typing.Optional[typing.List[typing.Union[Binding, typing.Tuple[str, str, str]]]]


class App(TextualApp):
    """Controlling Textual apps general behaviors"""

    def set_bindings(self, bindings: Bindings = None) -> None:
        """Sets the current bindings"""
        bindings = bindings or []
        self._bindings = _Bindings(bindings)



"""
The japanese translation

Copyright
---------
Animenosekai
    Original Author, MIT License
"""

from nasse.localization.base import Localization


class JapaneseLocalization(Localization):
    """
    The japanese version of the docs generation
    """
    __native__ = "日本語"

    sections = "セクション"
    getting_started = "はじめに"

    yes = "はい"
    no = "いいえ"

    no_description = "詳細なし"
    using_method = "{method}を使って"

    # Authentication
    authentication = "認証"
    no_auth_rule = "認証ルールが定義されていません"
    no_login = "ログインは**不要**です"
    login_with_types_required = "{types} アカウントでのログインが**必要**です"
    login_with_types_optional = "{types} アカウントでのログインは**任意**です"
    login_required = "ログインは**必要**です"
    login_optional = "ログインは**任意**です"
    login_suffix_only_verified = " （検証のみ）"

    # User sent values
    parameters = "パラメーター"
    headers = "ヘッダー"
    cookies = "クッキー"
    dynamic_url = "ダイナミックURL"

    name = "名前"
    description = "詳細"
    required = "必要"
    type = "タイプ"

    # Example
    example = "例"

    # Response
    response = "レスポンス"  # 応答や解答より「レスポンス」の方が使われているよう
    example_response = "レスポンスの例"
    not_json_response = "このエンドポイントは、JSON形式のレスポンスを返さないようです。"

    # Response description
    returns = "返したフィールド"
    field = "フィールド"
    nullable = "null可能"

    # Errors
    possible_errors = "起こりうるエラー"
    exception = "エラー名"
    code = "コード"

    # Index
    index = "インデックス"
    return_to_index = "インデックスに戻る"

    # Postman
    postman_description = "{name} APIインターフェイスでの「{section}」セクションの全てのエンドポイント"

    # Headers
    getting_started_header = '''
# {name} APIリファレンス

{name} APIリファレンスへようこそ！

## 全体的に

### レスポンスの形式

一般的に、JSONで形式されたレスポンスは以下の形に返されます（エラーが発生された時もこの形で返されます）

```json
{{
    "success": true,
    "message": "出来ました!",
    "error": null,
    "data": {{}}
}}
```

| フィールド     | 詳細                            |  null可能         |
| ------------ | ------------------------------  | ---------------- |
| `success`    | リクエストが成功したかどうか         | いいえ            |
| `message`    | 何があったか説明する文章            | はい             |
| `error`      | エラーが発生した時のエラー名         | はい             |
| `data`       | 解答のデータ、リクエストされたデータ  | いいえ            |

### エラー

サーバー上でも、リクエストからでも、様々なエラーが発生する可能性があります。

特定なエラーはそれぞれのエンドポイントドキュメントで説明されていなすが、以下に一般的に発生可能のエラーを説明します

| エラー名                     | 詳細                                                                                                             | コード  |
| --------------------------- | --------------------------------------------------------------------------------------------------------------- | ----- |
| `SERVER_ERROR`              | リクエストの処理中でサーバー上のエラーが発生した時                                                                       | 500   |
| `MISSING_CONTEXT`           | Nasseコンテキストに居ないのに、Nasseコンテキストのみで利用可能のデータをアクセスしようとしたとき                               | 500   |
| `INTERNAL_SERVER_ERROR`     | システム上でエラーが発生した時                                                                                       | 500   |
| `METHOD_NOT_ALLOWED`        | HTTPリクエストメソッドを間違えた時                                                                                   | 405   |
| `CLIENT_ERROR`              | リクエストで何か足りないかダメな時                                                                                    | 400   |
| `INVALID_TYPE`              | Nasseが送られたデータを正しい型に変換できなかった時                                                                     | 400   |
| `MISSING_VALUE`             | リクエストから何か足りない時                                                                                         | 400   |
| `MISSING_PARAM`             | 必要のパラメーターが一つリクエストから不足している時                                                                     | 400   |
| `MISSING_DYNAMIC`           | ダイナミックURLのパーツが一つURLから不足している時                                                                      | 400   |
| `MISSING_HEADER`            | 必要のヘッダーが一つリクエストから不足している時                                                                        | 400   |
| `MISSING_COOKIE`            | 必要のクッキーが一つリクエストから不足している時                                                                        | 400   |
| `AUTH_ERROR`                | 認証中にエラーが発生した時                                                                                           | 403   |

### 認証されたリクエスト

ユーザーがログインされていること

ユーザーのログインが必要な場合、「Authorization」ヘッダーがログインの時に送られたトークンに設定されることが必要です。

「{id}_token」のパラメーターと「__{id}_token」のクッキーも利用可能ですがこれらは優先されません

エンドポイントの認証ルールが「検証のみ」になっている場合は、アカウントがデータベースから読み込む事なく、トークンの形式や有効期限をチェックするだけです。

### デバッグモード

デバッグモードが有効の時(`-d`か`--debug`)は、色々な情報が`debug`フィールドで返送されて、`DEBUG`のロギングレベルが選ばれます。

'debug'フィルーどは重大なエラー（`INTERNAL_SERVER_ERROR`、 `METHOD_NOT_ALLOWED`、 など）しない限り、送られます。(重大なエラーが発生したら、なるべくもう一度エラーが発生しないようにする必要があります)

`call_stack`フィールドは`call_stack`パラメーターをリクエストで入れる時だけ送信されます。

```json
{{
    "success": true,
    "message": "リクエストが成功しませんでした",
    "error": null,
    "data": {{
        "username": "Animenosekai"
    }},
    "debug": {{
        "time": {{
            "global": 0.036757,
            "verification": 0.033558,
            "authentication": 0.003031,
            "processing": 4.9e-05,
            "formatting": 0.0001
        }},
        "ip": "127.0.0.1",
        "headers": {{
            "Host": "api.{id}.com",
            "Connection": "close",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Language": "fr-fr",
            "Accept-Encoding": "gzip, deflate, br",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
        }},
        "values": {{}},
        "domain": "api.{id}.com",
        "logs": [
            "1636562693.036563｜[INFO] [nasse.receive.Receive.__call__] → Incoming GET request to /account/name from 127.0.0.1",
            "1636562693.070008｜[ERROR] [nasse.exceptions.base.MissingToken.__init__] An authentication token is missing from the request"
        ],
        "call_stack": [
            "pass the 'call_stack' parameter to get the call stack"
        ]
    }}
}}
```

'''

    section_header = '''
# {name} セクションのAPIリファレンス

このファイルは「{name}」セクションの全てのエンドポイントを説明します。
'''

    # TUI

    # Footer
    tui_history = "履歴"
    tui_result = "結果"
    tui_explorer = "エクスプローラー"
    tui_submit = "送信"
    tui_options = "設定"
    tui_quit = "終了"

    # History
    # WARNING: This should stay under 3 characters to avoid having styling issues
    tui_min = "低"
    tui_average = "平均"
    tui_max = "高"

    # Explorer
    tui_reset = "リセット"

    # Request
    tui_request = "リクエスト"
    tui_name = "名前"
    tui_value = "内容"
    tui_path = "道"
    # tui_parameters = "Parameters"
    # tui_headers = "Headers"
    # tui_cookies = "Cookies"
    tui_file = "ファイル"
    tui_add_file = "ファイルを追加"
    tui_data = "データ"
    tui_add_data_file = "データファイルを追加"

    # Options
    tui_language = "言語"
    tui_language_notice = "言語変化を有効するには再起動して下さい"
    tui_base_url = "ベースURL"
    tui_base_url_placeholder = "リクエストとエクスプローラーのベースURL"
    tui_endpoints_update = "エンドポイントの更新度"
    tui_endpoints_update_placeholder = "エンドポイントの更新までの時間 (秒)"
    tui_history_limit = "履歴制限"
    tui_history_limit_placeholder = "履歴に残る最高のリクエスト数"
    tui_timeout = "タイムアウト"
    tui_timeout_placeholder = "タイムアウト (秒)"
    tui_redirects = "リダイレクト"
    tui_allow_redirects = "リダイレクトの許可"
    tui_proxies = "プロキシ"
    tui_security = "セキュリティー"
    tui_verify_request = "リクエストの確認"
    tui_certificate_files = "認証ファイル"
    tui_add_certificate = "認証ファイルの追加"

    # Result
    tui_start_prompt = "まずはリクエストを送信して下さい"
    tui_content = "内容"
    tui_no_content = "内容を表示できません"
    tui_contacting = "{url}を連絡しています"
    tui_files = "ファイル"
    tui_error = "エラー"

    # File Explorer
    tui_filter = "フィルター"

    # Quit
    tui_quit_confirmation = "本当に終了しますか？"
    tui_cancel = "戻る"



"""
This is where the requests are received and first processed
"""
import base64
import inspect
import sys
import typing

import flask

from nasse import config, exceptions, models, request, utils
from nasse.response import Response, exception_to_response
from nasse.utils import timer

RECEIVERS_COUNT = 0


def retrieve_token(context: request.Request = None) -> str:
    """
    Internal function to retrieve the login token from a request

    Parameters
    ----------
    context: nasse.request.request.Request
        The current request, if not properly set, the current context is used.
    """
    if not isinstance(context, request.Request):
        context = flask.g.request or flask.request
    token = context.headers.get("Authorization", None)
    if token is None:
        # should be app.id + "_token"
        token = context.values.get("{id}_token".format(id=context.nasse.config.id), None)
        if token is None:
            # should be "__" + app.id + "_token"
            token = context.cookies.get(
                "__{id}_token".format(id=context.nasse.config.id), None)
            if token is None:
                raise exceptions.authentication.MissingToken("An authentication token is missing from the request")
    return str(token)


class Receive:
    """The object which receives a request from Flask"""

    def __init__(self, app, endpoint: models.Endpoint) -> None:
        """
        This object is called by Flask and receives all of the requests.

        It performs some verification, according to what the user provided for the endpoint
        and sets some important variables, like nasse.request
        """
        # boo i need to use a global variable
        global RECEIVERS_COUNT
        self.app = app
        self.endpoint = endpoint
        RECEIVERS_COUNT += 1
        self.__name__ = "__nasse_receiver_{number}".format(number=RECEIVERS_COUNT)

    def __call__(self, *args: typing.Any, **kwds: typing.Any) -> typing.Any:
        try:
            with self.app.config.logger as logger:
                with utils.logging.CallStackRecorder() as call_stack:
                    with timer.Timer() as global_timer:
                        try:
                            with timer.Timer() as verification_timer:
                                context = request.Request(app=self.app, endpoint=self.endpoint, dynamics=kwds)
                                flask.g.request = context
                                logger.info("→ Incoming {{blue}}{method}{{normal}} request to {{blue}}{route}{{normal}} from {client}".format(method=flask.g.request.method,
                                                                                                                                              route=self.endpoint.path,
                                                                                                                                              client=flask.g.request.client_ip))

                            account = None
                            with timer.Timer() as authentication_timer:
                                method = flask.g.request.method
                                login_rules = self.endpoint.login.get(method, self.endpoint.login["*"])
                                for rule in login_rules:
                                    if not rule.skip:
                                        try:
                                            token = retrieve_token(context)
                                            if self.app.config.account_management:
                                                if not rule.skip_fetch:
                                                    account = self.app.config.account_management.retrieve_account(token)
                                                    if len(rule.types) > 0:
                                                        if self.app.config.account_management.retrieve_type(account) not in login_rules.types:
                                                            account = None  # if login is not required, the account might be passed with a wrong type
                                                            raise exceptions.authentication.Forbidden(
                                                                "You can't access this endpoint with your account")
                                                else:
                                                    verification = self.app.config.account_management.verify_token(token)
                                                    if verification == False:
                                                        raise exceptions.authentication.Forbidden("We couldn't verify your token")
                                            else:
                                                logger.warn("Couldn't verify login details because the 'account_management' is not set properly on {name}"
                                                            .format(name=self.app.config.name))
                                        except Exception as e:
                                            if rule.required:
                                                raise e

                            with timer.Timer() as processing_timer:
                                # TODO: Maybe use function signatures here ?
                                specs = inspect.getfullargspec(self.endpoint.handler)
                                arguments = {}

                                for arg in specs.args:  # for the function arguments
                                    for storage in (flask.g.request.values, flask.g.request.headers, flask.g.request.cookies):
                                        if arg in storage:
                                            val = storage.getlist(arg)
                                            if len(val) > 1:
                                                arguments[arg] = val
                                            else:
                                                arguments[arg] = val[0]
                                            break

                                if account is not None:
                                    arguments.pop("account", None)

                                for attr, current_values in [
                                    ("app", self.app),
                                    ("nasse", self.app),
                                    ("config", self.app.config),
                                    ("logger", self.app.logger),
                                    ("endpoint", self.endpoint),
                                    ("nasse_endpoint", self.endpoint),
                                    ("request", flask.g.request),
                                    ("method", flask.g.request.method),
                                    ("values", flask.g.request.values),
                                    ("params", flask.g.request.values),
                                    ("parameters", flask.g.request.values),
                                    ("args", flask.g.request.args),
                                    ("form", flask.g.request.form),
                                    ("headers", flask.g.request.headers),
                                    ("account", account),
                                    ("dynamics", flask.g.request.dynamics)
                                ]:
                                    if (attr in specs.args or specs.varkw) and attr not in arguments:
                                        arguments[attr] = current_values

                                for key, val in flask.g.request.dynamics.items():  # no need for multi=True as dynamics should only have one value
                                    if key in specs.args or specs.varkw:
                                        arguments[key] = val

                                # we don't need this because "dynamics" already holds those values
                                # and it saves us from making another loop
                                # arguments.update(kwds)
                                # for arg in arguments:
                                #     if arg not in specs:
                                #         arguments.pop(arg, None)

                                # calling the request handler
                                response = self.endpoint.handler(*args, **arguments)

                            with timer.Timer() as formatting_timer:
                                if isinstance(response, flask.Response):
                                    return response

                                data = None
                                error = None
                                message = None
                                code = 200
                                headers = {}
                                cookies = []

                                if isinstance(response, Response):
                                    # return Response(data=..., code=..., etc.)
                                    data = response.data
                                    code = response.code
                                    message = response.message
                                    error = response.error
                                    headers = response.headers
                                    cookies = response.cookies
                                elif isinstance(response, str):
                                    # return "Hello world"
                                    message = response
                                elif isinstance(response, bytes):
                                    # return b"binary data"
                                    data = response
                                elif isinstance(response, Exception):
                                    # return NasseException("Something went wrong")
                                    message, error, code = exception_to_response(response)
                                elif isinstance(response, typing.Iterable) and not isinstance(response, typing.Generator):
                                    found = False
                                    if utils.unpack.is_unpackable(response):
                                        try:
                                            response = Response(**response)
                                            data = response.data
                                            code = response.code
                                            error = response.error
                                            headers = response.headers
                                            cookies = response.cookies
                                            message = response.message
                                            found = True
                                        except TypeError:
                                            data = response
                                    if not found:
                                        # return "Hello", 200 | return NasseException("..."), 400, {"extra": {"issue": "something is missing"}}
                                        for value in response:
                                            if isinstance(value, Exception):
                                                message, error, code = exception_to_response(value)
                                            elif isinstance(value, int):
                                                code = int(value)
                                            elif isinstance(value, str):
                                                message = value
                                            else:
                                                data = value
                                else:
                                    data = response

                                if code < 100 or code >= 600:
                                    logger.warn("The returning HTTP status code doesn't seem to be a standard status code: {code}"
                                                .format(code=code))

                                if not self.endpoint.json:
                                    final = flask.Response(response=data, status=code)

                                    try:
                                        if error:
                                            final.headers["X-NASSE-ERROR"] = str(error)

                                        if self.app.debug:
                                            final.headers["X-NASSE-TIME-GLOBAL"] = str(global_timer.stop())
                                            final.headers["X-NASSE-TIME-VERIFICATION"] = str(verification_timer.time)
                                            final.headers["X-NASSE-TIME-AUTHENTICATION"] = str(authentication_timer.time)
                                            final.headers["X-NASSE-TIME-PROCESSING"] = str(processing_timer.time)
                                            final.headers["X-NASSE-TIME-FORMATTING"] = str(formatting_timer.stop())
                                    except Exception:
                                        pass
                                else:
                                    result = {
                                        "success": error is None,
                                        "error": error,
                                        "message": message,
                                        "data": {}
                                    }
                                    if utils.unpack.is_unpackable(data):
                                        # data: {"username": "someone", "token": "something"}
                                        result["data"] = dict(data)
                                    elif isinstance(data, bytes):
                                        # data: bytes data, raw file content
                                        result["data"]["base64"] = base64.b64encode(data).decode("utf-8")
                                    elif hasattr(data, "read") and hasattr(data, "tell") and hasattr(data, "seek"):
                                        # a file object
                                        position = data.tell()  # storing the current position
                                        content = data.read()  # read it (place the cursor at the end)
                                        # go back to the original position
                                        data.seek(position)
                                        if "b" in data.mode:  # if binary mode
                                            result["data"]["base64"] = base64.b64encode(content).decode("utf-8")  # base64 encode the result
                                        else:
                                            result["data"]["content"] = str(content)
                                    elif isinstance(data, str):
                                        # data: "Hello World"
                                        result["data"]["string"] = data
                                    elif isinstance(data, typing.Iterable):
                                        # data: ["an", "array", "of", "element"] | ("an", "array", ...) | etc.
                                        result["data"]["array"] = list(data)
                                    elif data is None:
                                        result["data"] = None
                                    else:
                                        # data: typing.Any (but json does not support arbitrary content)
                                        logger.warn("Element of type <{type}> is not supported by JSON and will be converted to `str`".format(
                                            type=data.__class__.__name__))
                                        result["data"]["content"] = str(data)
                        except Exception as e:
                            if self.app.config.debug:
                                if not isinstance(e, exceptions.authentication.MissingToken):
                                    logger.print_exception()
                                else:
                                    logger.debug("The request seems to be lacking an authentication token")

                            message, error, code = exception_to_response(e)
                            result = {
                                "success": False,
                                "error": error,
                                "message": message,
                                "data": {}
                            }

                        try:
                            flask.g.request
                        except Exception:
                            flask.g.request = flask.request

                        try:
                            verification_timer
                        except Exception:
                            verification_timer = None
                        try:
                            authentication_timer
                        except Exception:
                            authentication_timer = None
                        try:
                            processing_timer
                        except Exception:
                            processing_timer = None
                        try:
                            formatting_timer
                        except Exception:
                            formatting_timer = None

                        try:
                            result
                        except Exception:
                            result = {}

                        if self.endpoint.json:
                            try:
                                flask.g.request
                            except Exception:
                                flask.g.request = flask.request

                            if self.app.config.debug:
                                result["debug"] = {
                                    "time": {
                                        "global": global_timer.stop(),
                                        "verification": verification_timer.time if verification_timer else None,
                                        "authentication": authentication_timer.time if authentication_timer else None,
                                        "processing": processing_timer.time if processing_timer else None,
                                        "formatting": formatting_timer.stop() if formatting_timer else None
                                    },
                                    "ip": flask.g.request.client_ip if isinstance(flask.g.request, request.Request) else utils.ip.get_ip(),
                                    "headers": dict(flask.g.request.headers),
                                    "values": dict(flask.g.request.values),
                                    "domain": flask.g.request.host,
                                    "logs": [{
                                        "time": r.time,
                                        "level": r.level.name,
                                        "msg": r.msg
                                    } for r in logger.record],
                                    "call_stack": ["pass the 'call_stack' parameter to get the call stack"]
                                }

                                if "call_stack" in flask.g.request.values:
                                    result["debug"]["call_stack"] = [frame.as_dict()
                                                                     for frame in call_stack.call_stack]

                            minify = utils.boolean.to_bool(flask.g.request.values.get("minify", False))

                            content_type = "application/json"
                            if utils.sanitize.remove_spaces(flask.g.request.values.get("format", "json")).lower() in {"xml", "html"}:
                                body = utils.xml.encode(data=result, minify=minify)
                                content_type = "application/xml"
                            else:
                                body = (utils.json.minified_encoder if minify else utils.json.encoder).encode(result)

                            final = flask.Response(body, status=code)
                            final.headers["Content-Type"] = content_type

                try:
                    final
                except Exception:
                    # final is not defined, meaning that the exception was raised before the response was created
                    final = flask.Response(None, status=code)

                    try:
                        if error:
                            final.headers["X-NASSE-ERROR"] = str(error)

                        if self.app.config.debug:
                            final.headers["X-NASSE-TIME-GLOBAL"] = str(global_timer.stop())
                            final.headers["X-NASSE-TIME-VERIFICATION"] = str(verification_timer.time) if verification_timer is not None else "N/A"
                            final.headers["X-NASSE-TIME-AUTHENTICATION"] = str(
                                authentication_timer.time) if authentication_timer is not None else "N/A"
                            final.headers["X-NASSE-TIME-PROCESSING"] = str(processing_timer.time) if processing_timer is not None else "N/A"
                            final.headers["X-NASSE-TIME-FORMATTING"] = str(formatting_timer.stop()) if formatting_timer is not None else "N/A"
                    except Exception:
                        pass

                try:
                    flask.g.request
                except Exception:
                    flask.g.request = flask.request
                try:
                    cookies
                except Exception:
                    cookies = []
                try:
                    headers
                except Exception:
                    headers = {}

                # final is now defined
                for cookie in cookies:
                    # 'cookie' should be a ResponseCookie
                    final.headers.add("Set-Cookie", cookie.dumps())

                for key, value in headers.items():
                    final.headers.add(str(key), str(value))

                try:
                    size = sys.getsizeof(final.data)
                    if isinstance(flask.g.request, request.Request):
                        path = flask.g.request.nasse_endpoint.path
                        ip = flask.g.request.client_ip
                        method = flask.g.request.method
                    else:
                        path = flask.g.request.path
                        ip = utils.ip.get_ip()
                        method = str(flask.g.request.method).upper()
                    if size < 500000:
                        color = "{green}"
                    elif size < 1000000:
                        color = "{yellow}"
                    else:
                        color = "{magenta}"
                    if final.status_code < 200:  # ?
                        status_color = "{white}"
                    elif final.status_code < 300:
                        status_color = "{blue}"
                    elif final.status_code < 400:
                        status_color = "{green}"
                    elif final.status_code < 500:
                        status_color = "{yellow}"
                    else:
                        status_color = "{magenta}"
                    self.app.config.logger.info("← Sending back {color}{size}{{normal}} bytes of data to {ip} following {status_color}{status}{{normal}} {{blue}}{method} {path}{{normal}}".format(
                        color=color,
                        size=size,
                        ip=ip,
                        method=method,
                        status=final.status_code,
                        status_color=status_color,
                        path=path,
                    ))
                except Exception:
                    # utils.logging.logger.print_exception()
                    pass

                return final
        except Exception as err:
            utils.logging.logger.print_exception(show_locals=True)
            raise err



"""Defines the options view"""
import dataclasses
import json
import pathlib
import typing

from textual import events
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button

from nasse.tui.components.headers import StickyHeader

T = typing.TypeVar("T")


class OptionsScreen(ModalScreen[T]):
    """The options managing screen"""

    DEFAULT_CSS = """
    OptionsScreen {
        align: center middle;
        background: rgba(30, 30, 30, 0.75);
        height: auto;
        width: 80vw;
    }

    Input {
        margin-bottom: 1;
    }

    .options-switch-title {
        content-align: center middle;
        height: 3;
        content-align: center middle;
        width: 20;
    }

    .options-switch-container {
        align-vertical: middle;
        margin-bottom: 1;
        height: auto;
        width: auto;
    }

    #options-container {
        /* height: auto; */
        min-height: 50vh;
        max-height: 75vh;
        width: 80vw;
        height: auto;
        padding: 1 5;
        content-align: center middle;
        align: center middle;
        border: round gray;
    }

    #options-confirmation-container {
        width: auto;
        height: auto;
        dock: bottom;
        align-horizontal: right;
    }
    """

    def __init__(self,
                 options: T,
                 name: typing.Optional[str] = None,
                 id: typing.Optional[str] = None,
                 classes: typing.Optional[str] = None) -> None:
        super().__init__(name, id, classes)
        self.options: T = options

    def compose(self):
        yield StickyHeader("Options")
        with Container(id="options-container"):
            yield from self.compose_options()
        with Horizontal(id="options-confirmation-container"):
            yield Button("Ok", id="options-confirmation-button")

    def compose_options(self):
        """The inner options view renderer"""
        raise NotImplementedError("The options view couldn't be rendered because it is not implemented yet.")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """When a button is pressed"""
        if event.button.id == "options-confirmation-button":
            options = dataclasses.asdict(self.options)
            options.update(self.collect_values())
            return self.dismiss(self.options.__class__(**options))

    def collect_values(self) -> typing.Dict[str, typing.Any]:
        """Collect the different options value"""
        return {}

    def on_key(self, event: events.Key) -> None:
        """When a key is pressed on the keyboard"""
        if event.key == "escape":
            return self.dismiss(self.options)

    @staticmethod
    def loads(key: str, cast: typing.Type[T]) -> T:
        """Loads the configs"""
        config_path = pathlib.Path() / ".nasse" / "config" / str(key)
        try:
            return cast(**json.loads(config_path.read_text()))
        except Exception:
            return cast()

    @staticmethod
    def dumps(key: str, options: T) -> None:
        """Exports the configs"""
        config_path = pathlib.Path() / ".nasse" / "config" / str(key)
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.write_text(json.dumps(dataclasses.asdict(options), ensure_ascii=False, separators=(",", ":")))



"""A file explorer"""
import pathlib
import typing

from textual import events
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import DirectoryTree, Label, Tree
from textual.widgets._directory_tree import DirEntry
from nasse.localization import Localization, EnglishLocalization


class FilteredDirectoryTree(DirectoryTree):
    """A filtered directory tree"""
    filter: reactive[typing.Optional[str]] = reactive(None)

    def filter_paths(self, paths: typing.Iterable[pathlib.Path]) -> typing.Iterable[pathlib.Path]:
        if not self.filter:
            return paths
        return [path for path in paths if str(path.resolve()).casefold().startswith(self.filter) or path.name.casefold().startswith(self.filter)]

    def watch_filter(self, filter: typing.Optional[str]):
        """When the filter is changed"""
        self.reload()


class FileBrowser(ModalScreen[typing.Optional[pathlib.Path]]):
    """A file browser"""
    selected: reactive[pathlib.Path] = reactive(pathlib.Path)

    DEFAULT_CSS = """
    FileBrowser {
        height: 80vh;
        width: 80vw;
    }
    #filter-label {
        opacity: 1;
        height: auto;
        background: $panel-darken-1;
        width: 100vw;
        padding: 0 1;
    }

    #filter-label.hidden {
        opacity: 0;
    }
    """
    def __init__(self, localization: typing.Type[Localization] = EnglishLocalization, **kwargs) -> None:
        super().__init__(**kwargs)
        self.localization = localization

    def compose(self):
        yield Label(str(pathlib.Path().resolve()), id="file-browser-title")
        yield FilteredDirectoryTree(pathlib.Path())
        yield Label("", id="filter-label")

    def go_back(self) -> None:
        """When the user wants to go back"""
        tree = self.query_one(FilteredDirectoryTree)
        tree.path = pathlib.Path(tree.path).resolve().parent

    def go_forward(self) -> None:
        """When the user wants to go back"""
        tree = self.query_one(FilteredDirectoryTree)
        path = self.selected.resolve()
        if path.is_dir():
            tree.path = path

    def on_tree_node_highlighted(self, event: Tree.NodeHighlighted[DirEntry]):
        """When a node is highlited"""
        if not event.node.data:
            return
        path = event.node.data.path.resolve()
        self.selected = path

    def watch_selected(self, selected: pathlib.Path):
        """When the selection changed"""
        parent, sep, filename = str(selected).rpartition("/")

        if not filename:
            parent, sep, filename = str(selected).rpartition("\\")

        self.query_one("#file-browser-title", Label).update(f"[grey]{parent}{sep}[/grey][bold]{filename}[/bold]")

    def on_directory_tree_file_selected(self, event: DirectoryTree.FileSelected):
        """When a file is selected"""
        if self.selected.is_file():
            self.dismiss(self.selected)

    def on_key(self, event: events.Key) -> None:
        """When a key is pressed on the keyboard"""
        directory_tree = self.query_one(FilteredDirectoryTree)
        if event.key == "escape":
            if not directory_tree.filter:
                self.dismiss(None)
            else:
                directory_tree.filter = None
        elif event.key == "delete" or event.key == "backspace":
            if directory_tree.filter:
                directory_tree.filter = directory_tree.filter[:-1]
        elif event.key == "left":
            self.go_back()
        elif event.key == "right":
            self.go_forward()
        elif event.is_printable and event.character:
            if directory_tree.filter:
                directory_tree.filter += event.character.casefold()
            else:
                directory_tree.filter = event.character.casefold()

        label = self.query_one("#filter-label", Label)
        if directory_tree.filter:
            label.update(f"[grey]{self.localization.tui_filter}:[/grey] [bold]{directory_tree.filter}[/bold]")
            label.remove_class("hidden")
        else:
            label.add_class("hidden")



"""
Helper to create static Python code from Nasse objects
"""

import pathlib
import json
from nasse.models import Endpoint, Error, Login, Return, UserSent
# from nasse.utils.logging import logger


def return_to_python(element: Return,
                     indent: str = " " * 4,
                     end_indent: str = "") -> str:
    """Turns the given `Return` object to a static python snippet of code"""
    result = "Return(\n"
    result += indent + "name=" + \
        json.dumps(element.name, ensure_ascii=False) + ",\n"
    if element.example:
        result += indent + "example=" + (json.dumps(element.example, ensure_ascii=False)
                                           if isinstance(element.example, str)
                                           else str(element.example)) + ",\n"
    if element.description:
        result += indent + "description=" + \
            json.dumps(element.description, ensure_ascii=False) + ",\n"
    if element.type:
        result += indent + "type=" + \
            json.dumps(str(element.type), ensure_ascii=False) + ",\n"
    if element.children:
        result += indent + "children=" + str(element.children) + ",\n"
    result += indent + "nullable = " + str(element.nullable) + "\n"
    result += end_indent + ")"
    return result


def login_to_python(element: Login,
                    indent: str = " " * 4,
                    end_indent: str = "") -> str:
    """Turns the given `Login` object to a static python snippet of code"""
    result = "Login(\n"
    if element.required:
        result += indent + "required=True,\n"
    if element.types:
        # we can't really convert non-str types
        result += indent + "types=" + str(element.types) + ",\n"
    if element.skip:
        result += indent + "skip=True,\n"
    if element.skip_fetch:
        result += indent + "skip_fetch=True\n"

    result += end_indent + ")"
    return result


def usersent_to_python(element: UserSent,
                       indent: str = " " * 4,
                       end_indent: str = "") -> str:
    """Turns the given `UserSent` object to a static python snippet of code"""
    result = "{t}(\n".format(t=element.__class__.__name__)
    result += indent + "name=" + \
        json.dumps(element.name, ensure_ascii=False) + ",\n"

    if element.description:
        result += indent + "description=" + \
            json.dumps(element.description, ensure_ascii=False) + ",\n"
    if not element.required:
        result += indent + "required=False,\n"

    if element.type != str:
        has_name = (hasattr(element, '__name__') and not isinstance(element, str))
        result += indent + f"type={element.type.__name__ if has_name else str(element.type)}"

    # we can't convert types for now
    # logger.warn("It is not possible to convert types to static python code for now")

    result += end_indent + ")"
    return result


def error_to_python(element: Error,
                    indent: str = " " * 4,
                    end_indent: str = "") -> str:
    """Turns the given `Error` object to a static python snippet of code"""
    result = "Error(\n"
    result += indent + "name=" + \
        json.dumps(element.name, ensure_ascii=False) + ",\n"

    if element.description:
        result += indent + "description=" + \
            json.dumps(element.description, ensure_ascii=False) + ",\n"

    if element.code != 500:
        result += indent + "code=" + str(element.code) + ",\n"

    result += end_indent + ")"
    return result


def endpoint_to_python(endpoint: Endpoint, explicit_path: bool = True, indent: str = " " * 4) -> str:
    """Turns the given `Endpoint` object to a static python snippet of code"""
    result = "Endpoint(\n"

    if explicit_path:
        result += indent + "path=" + \
            json.dumps(endpoint.path, ensure_ascii=False) + ",\n"

    if not len(endpoint.methods) == 1 or list(endpoint.methods)[0] != "GET":
        result += indent + "methods=" + str(endpoint.methods) + ",\n"

    if not endpoint.json:
        result += indent + "json=False"

    for attribute in ("name", "category", "sub_category", "description", "base_dir"):
        element = getattr(endpoint, attribute)
        if isinstance(element, pathlib.Path):
            element = str(element)
        if isinstance(element, set):
            element = list(element)
        result += indent + f"{attribute}=" + \
            json.dumps(element, ensure_ascii=False) + ",\n"

    for attribute, caster in [("headers", usersent_to_python),
                              ("cookies", usersent_to_python),
                              ("parameters", usersent_to_python),
                              ("dynamics", usersent_to_python),
                              ("returns", return_to_python),
                              ("login", login_to_python),
                              ("errors", error_to_python)]:
        element = getattr(endpoint, attribute)
        if not element or (len(element) == 1 and
                           "*" in element and
                           not element.get("*", set())):
            continue

        result += indent + attribute + "={\n"
        for method, values in element.items():
            if not values:
                continue
            result += indent * 2 + f'"{method}": [\n'
            for element in values:
                result += indent * 3 + \
                    caster(element, indent=indent * 4,
                           end_indent=indent * 3) + ",\n"
            result += indent * 2 + "],\n"
        result += indent + "},\n"

    result += ")"
    return result


def endpoint_to_decorator(endpoint: Endpoint, app_name: str = "app", explicit_path: bool = True, indent: str = " " * 4):
    """Turns the given `Endpoint` object to a static python snippet of code with a decorator"""
    return "@{name}.route(endpoint={result})".format(name=app_name,
                                            result=endpoint_to_python(endpoint=endpoint, explicit_path=explicit_path, indent=indent))



"""
Generates response examples
"""

from nasse import models, utils


def _get_type(data: models.Return) -> str:
    """
    An internal function to retrieves the type of the given returning item

    Parameters
    ----------
    data: models.Return
        The return value to get the type from

    Returns
    -------
    "string" | "int" | "float | "bool" | "array" | "object"
        The type, when possible
    str
        When impossible, returns the type as a string
    """
    key_type = data.type

    if key_type is None:
        example = data.example
        if example is None:
            return "string"
        if isinstance(example, type):
            return example.__name__
        elif isinstance(example, list):
            return "array"
        elif isinstance(example, dict):
            return "object"
        else:
            try:
                return example.__class__.__name__  # should be `str`
            except AttributeError:
                pass

    key_type = str(key_type)
    key_type_token = key_type.lower().replace(" ", "")
    if key_type_token in {"str", "string", "text"}:
        return "string"
    elif key_type_token in {"int", "integer"}:
        return "int"
    elif key_type_token in {"float", "floating", "number"}:
        return "float"
    elif key_type_token in {"bool", "true", "false", "boolean"}:
        return "bool"
    elif key_type_token in {"array", "list", "arr", "set"}:
        return "array"
    elif key_type_token in {"object", "obj", "dict", "dictionary"}:
        return "object"
    return key_type


def generate_example(endpoint: models.Endpoint, method: str) -> str:
    """
    Generates a JSON response example

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to get the response example from
    method: str
        The method to get the response example from

    Returns
    -------
    str
        The response example
    """
    def get_value(data):
        """

        Parameters
        ----------
        data
        """
        key_type = _get_type(data)
        # if self.returning[key].get("nullable", False) and key_type != "object":
        #    return None
        if key_type == "str":
            return str(data.example or "string")
        elif key_type == "int":
            return int(data.example or 4)
        elif key_type == "float":
            return float(data.example or 4.6)
        elif key_type == "array":
            return list(data.example or [])
        elif key_type == "bool":
            return utils.boolean.to_bool(data.example or True)
        elif key_type == "object":
            return {child.name: get_value(child) for child in data.children if (data.all_methods or method in data.methods)}
        return data.example or "no example"

    _response_format = {}
    for element in [data for data in models.get_method_variant(method, endpoint.returns) if "." not in data.name]:
        _response_format[element.name] = get_value(element)

    # we can't consider using f-strings because they came with py3.6
    # pylint: disable=consider-using-f-string
    return '''{{
    "success": true,
    "message": "Successfully processed your request",
    "error": null,
    "data": {response}
}}
'''.format(response=utils.json.encoder.encode(_response_format).replace("\n", "\n    "))



from nasse.localization.base import Localization
from nasse.localization.fra import FrenchLocalization
from nasse.localization.eng import EnglishLocalization
from nasse.localization.jpn import JapaneseLocalization



"""A route parser"""
import typing
from werkzeug.routing.converters import DEFAULT_CONVERTERS as FLASK_DEFAULT_CONVERTERS


class MismatchError(ValueError):
    """When the paths don't match"""


class PathNotFound(ValueError):
    """When we couldn't find the path"""


class CastError(ValueError):
    """When we couldn't cast the value to the given type"""


class Part:
    """Part of a Path"""

    def __init__(self, part: str) -> None:
        self.part = str(part)


class Dynamic(Part):
    """The dynamic part of a Path"""

    # pylint: disable=redefined-builtin
    def __init__(self, part: str, name: str, type: str = "str") -> None:
        super().__init__(part)

        self.name = str(name).strip()
        self.type = str(type).strip().lower()
        self.cast = self.caster(self.type)

        if self.type:
            self.part = f"<{self.type}:{self.name}>"
        else:
            self.part = f"<{self.name}>"

    @classmethod
    def caster(cls, type: str):
        """Returns the python class associated with the type"""
        type = str(type).lower()
        if type == "int":
            return int
        if type == "float":
            return float
        # could add more
        return str

    @property
    def flask(self) -> str:
        """Returns a flask compatible version of the part"""
        if self.type == "str":
            fixed_type = "string"
        else:
            fixed_type = self.type

        if fixed_type and fixed_type in FLASK_DEFAULT_CONVERTERS:
            return f"<{fixed_type}:{self.name}>"
        return f"<{self.name}>"


class Path:
    """A parsed Path"""

    def __init__(self, path: str) -> None:
        self.parts = self.parse(path)
        self.path = "/" + "/".join(element.part for element in self.parts)

        # Caching
        self._dynamics_num = len(self.dynamics)

    def __repr__(self) -> str:
        return f'Path("{self.path}")'

    @property
    def dynamics(self):
        """Returns the different dynamics parts of the path"""
        return [part for part in self.parts if isinstance(part, Dynamic)]

    def join(self, flask: bool = False) -> str:
        """Joins the different parts into a string"""
        if flask:
            return "/" + "/".join(element.flask if isinstance(element, Dynamic)
                                  else element.part
                                  for element in self.parts)
        return "/" + "/".join(element.part for element in self.parts)

    @classmethod
    def splitter(cls, path: str):
        """Splits the given path"""
        return str(path).strip("/").split("/")

    @classmethod
    def parse(cls, path: str) -> typing.List[Part]:
        """Returns the different parst for the given Path"""
        results = []
        for element in cls.splitter(path):
            element = str(element).strip()
            if element.startswith("<") and element.endswith(">"):
                # pylint: disable=redefined-builtin
                type, _, name = element.lstrip("<").rstrip(">").partition(":")
                if not name:
                    results.append(Dynamic(part=element, name=type))
                else:
                    results.append(Dynamic(
                        part=element,
                        name=name,
                        type=type
                    ))
            else:
                results.append(Part(part=element))
        return results

    def resolve(self, path: str):
        """Returns the different dynamic parts of the URL"""
        parts = self.splitter(path)
        if len(parts) != len(self.parts):
            raise MismatchError("The given path does not seem to be taking the right route")

        # Could use a list instead of a dictionary ?
        results = {}
        for index, part in enumerate(parts):
            element = self.parts[index]

            if isinstance(element, Dynamic):
                try:
                    results[element.name] = element.cast(part)
                except Exception as exc:
                    raise CastError(f"Couldn't cast to `{element.cast}`") from exc
                continue

            if element.part != part:
                raise MismatchError("The given path does not seem to be taking the right route")

        return results

    def is_route(self, path: str):
        """Returns if it is taking the right route"""
        try:
            self.resolve(path)
            return True
        except MismatchError:
            return False


def resolve(path: str, routes: typing.Iterable[Path]):
    """Resolves the path to determine which route to go for"""
    results: typing.List[typing.Tuple[Path, typing.Dict[str, typing.Any]]] = []

    for route in routes:
        try:
            results.append((route, route.resolve(path)))
        except MismatchError:
            continue

    if not results:
        raise PathNotFound("(404) Couldn't find the given path")

    # pylint: disable=protected-access
    return sorted(results, key=lambda element: element[0]._dynamics_num)[0]



"""
The base strings
"""
import pathlib
import inspect


# We are using a special `classproperty` because
# the chaining of `@property` and `@classmethod`
# got deprecated in 3.11
class classproperty(property):
    """A class property"""

    def __get__(self, owner_self, owner_cls):
        if self.fget:
            return self.fget(owner_cls)


class Localization:
    """
    Represents a Nasse documentation generation localization
    """
    __native__ = "Base"
    """The native name for the localization language"""

    @classproperty
    def __id__(cls):  # pylint: disable=no-self-argument
        return pathlib.Path(inspect.getfile(cls)).stem

    @classproperty
    def __english__(cls):  # pylint: disable=no-self-argument
        # Might change this later to support MultipleCaseNames
        # pylint: disable=no-member
        return cls.__name__.lower().removesuffix("localization").title()

    sections = "Sections"
    getting_started = "Getting Started"

    yes = "Yes"
    no = "No"

    no_description = "No description"
    using_method = "Using {method}"

    # Authentication
    authentication = "Authentication"
    no_auth_rule = "There is no authentication rule defined"
    no_login = "Login is **not** required"
    login_with_types_required = "Login with {types} is **required**"
    login_with_types_optional = "Login with {types} is **optional**"
    login_required = "Login is **required**"
    login_optional = "Login is **optional**"
    login_suffix_only_verified = " but only verified"

    # User sent values
    parameters = "Parameters"
    headers = "Headers"
    cookies = "Cookies"
    dynamic_url = "Dynamic URL"

    name = "Name"
    description = "Description"
    required = "Required"
    type = "Type"

    # Example
    example = "Example"

    # Response
    response = "Response"
    example_response = "Example response"
    not_json_response = "This endpoint doesn't seem to return a JSON-formatted response."

    # Response description
    returns = "Returns"
    field = "Field"
    nullable = "Nullable"

    # Errors
    possible_errors = "Possible Errors"
    exception = "Exception"
    code = "Code"

    # Index
    index = "Index"
    return_to_index = "Return to the Index"

    # Postman
    postman_description = "All of the endpoints under the '{section}' section of the {name} API Interface"

    # Headers
    getting_started_header = '''
# {name} API Reference

Welcome to the {name} API Reference.

## Globals

### Response Format

Globally, JSON responses should be formatted as follows (even when critical errors occur)

```json
{{
    "success": true,
    "message": "We successfully did this!",
    "error": null,
    "data": {{}}
}}
```

| Field        | Description                                      | Nullable         |
| ------------ | ------------------------------------------------ | ---------------- |
| `success`    | Whether the request was a success or not         | No               |
| `message`    | A message describing what happened               | Yes              |
| `error`      | The exception name if an error occurred          | Yes              |
| `data`       | The extra data, information asked in the request | No               |

### Errors

Multiple Errors can occur, server side or request side.

Specific errors are documented in each endpoint, but these are the general errors that can occur on any endpoint:

| Exception                   | Description                                                                                                     | Code  |
| --------------------------- | --------------------------------------------------------------------------------------------------------------- | ----- |
| `SERVER_ERROR`              | When an error occurs on {name} while processing a request                                                       | 500   |
| `MISSING_CONTEXT`           | When you are trying to access something which is only available in a Nasse context, and you aren't in one       | 500   |
| `INTERNAL_SERVER_ERROR`     | When a critical error occurs on the system                                                                      | 500   |
| `METHOD_NOT_ALLOWED`        | When you made a request with the wrong method                                                                   | 405   |
| `CLIENT_ERROR`              | When something is missing or is wrong with the request                                                          | 400   |
| `INVALID_TYPE`              | When Nasse couldn't convert the given value to the right type                                                   | 400   |
| `MISSING_VALUE`             | When a value is missing from the request                                                                        | 400   |
| `MISSING_PARAM`             | When a parameter is missing from the request                                                                    | 400   |
| `MISSING_DYNAMIC`           | When a dynamic routing value is missing from the requested URL                                                  | 400   |
| `MISSING_HEADER`            | When a header is missing from the request                                                                       | 400   |
| `MISSING_COOKIE`            | When a cookie is missing from the request                                                                       | 400   |
| `AUTH_ERROR`                | When an error occurred while authenticating the request                                                         | 403   |

### Authenticated Requests

When a user needs to be logged in, the "Authorization" header needs to be set to the login token provided when logging in.

Alternatively, the "{id}_token" parameter and "__{id}_token" cookie can be used, but these won't be prioritized.

If the endpoint is flagged for a "verified only" login, the account won't be fetched from any database, but the token will be checked.

### Debug Mode

On debug mode (`-d` or `--debug`), multiple information are passed in the `debug` section of the response and the `DEBUG` log level is selected on the server.

The 'debug' section is available on every type of error, except the ones issued by Flask such as `INTERNAL_SERVER_ERROR`, `METHOD_NOT_ALLOWED`, etc. (they need to do the bare minimum to not raise an exception and therefore breaking the server)

The "call_stack" attribute is enabled only when an error occurs or the `call_stack` parameter is passed with the request.

```json
{{
    "success": true,
    "message": "We couldn't fullfil your request",
    "error": null,
    "data": {{
        "username": "Animenosekai"
    }},
    "debug": {{
        "time": {{
            "global": 0.036757,
            "verification": 0.033558,
            "authentication": 0.003031,
            "processing": 4.9e-05,
            "formatting": 0.0001
        }},
        "ip": "127.0.0.1",
        "headers": {{
            "Host": "api.{id}.com",
            "Connection": "close",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Language": "fr-fr",
            "Accept-Encoding": "gzip, deflate, br",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
        }},
        "values": {{}},
        "domain": "api.{id}.com",
        "logs": [
            "1636562693.036563｜[INFO] [nasse.receive.Receive.__call__] → Incoming GET request to /account/name from 127.0.0.1",
            "1636562693.070008｜[ERROR] [nasse.exceptions.base.MissingToken.__init__] An authentication token is missing from the request"
        ],
        "call_stack": [
            "pass the 'call_stack' parameter to get the call stack"
        ]
    }}
}}
```

'''

    section_header = '''
# {name} Section API Reference

This file lists and explains the different endpoints available in the {name} section.
'''

    # TUI

    # Footer
    tui_history = "History"
    tui_result = "Result"
    tui_explorer = "Explorer"
    tui_submit = "Submit"
    tui_options = "Options"
    tui_quit = "Quit"

    # History
    # WARNING: This should stay under 3 characters to avoid having styling issues
    tui_min = "Min"
    tui_average = "Avg"
    tui_max = "Max"

    # Explorer
    tui_reset = "Reset"

    # Request
    tui_request = "Request"
    tui_name = "name"
    tui_value = "value"
    tui_path = "path"
    # tui_parameters = "Parameters"
    # tui_headers = "Headers"
    # tui_cookies = "Cookies"
    tui_file = "File"
    tui_add_file = "Add File"
    tui_data = "Data"
    tui_add_data_file = "Add Data File"

    # Options
    tui_language = "Language"
    tui_language_notice = "You need to restart the app to apply the changes"
    tui_base_url = "Base URL"
    tui_base_url_placeholder = "the base url for the requests and the endpoints explorer"
    tui_endpoints_update = "Endpoints Update"
    tui_endpoints_update_placeholder = "time between each endpoints list update (in sec.)"
    tui_history_limit = "History Limit"
    tui_history_limit_placeholder = "maximum number of requests in the history"
    tui_timeout = "Timeout"
    tui_timeout_placeholder = "timeout (sec.)"
    tui_redirects = "Redirects"
    tui_allow_redirects = "Allow Redirects"
    tui_proxies = "Proxies"
    tui_security = "Security"
    tui_verify_request = "Verify Request"
    tui_certificate_files = "Certificate Files"
    tui_add_certificate = "Add Certificate"

    # Result
    tui_start_prompt = "Start by making a request"
    tui_content = "Content"
    tui_no_content = "Can't display the content"
    tui_contacting = "Contacting {url}"
    tui_files = "Files"
    tui_error = "Error"

    # File Explorer
    tui_filter = "Filter"

    # Quit
    tui_quit_confirmation = "Are you sure you want to quit?"
    tui_cancel = "Cancel"



"""
Nasse's sanitizing and convert utility
"""

import typing

import nh3
from nasse import utils

# Source: en.wikipedia.org/wiki/Whitespace_character
# Note: BRAILLE PATTERN BLANK, HANGUL FILLER, HANGUL CHOSEONG FILLER, HANGUL JUNGSEONG FILLER and
# HALFWIDTH HANGUL FILLER are also refered here as "whitespaces" while they aren't according to the Unicode standard.
WHITESPACES = ["\u0009", "\u000A", "\u000B", "\u000C", "\u000D", "\u0020", "\u0085",
               "\u00A0", "\u1680", "\u2000", "\u2001", "\u2002", "\u2003", "\u2004",
               "\u2005", "\u2006", "\u2007", "\u2008", "\u2009", "\u200A", "\u2028",
               "\u2029", "\u202F", "\u205F", "\u3000", "\u180E", "\u200B", "\u200C",
               "\u200D", "\u2060", "\uFEFF", "\u00B7", "\u21A1", "\u2261", "\u237D",
               "\u23CE", "\u2409", "\u240A", "\u240B", "\u240C", "\u240D", "\u2420",
               "\u2422", "\u2423", "\u2424", "\u25B3", "\u2A5B", "\u2AAA", "\u2AAB",
               "\u3037", "\u2800", "\u3164", "\u115F", "\u1160", "\uFFA0"]
"""A list of whitespace characters"""

ALLOWED_TAGS = {'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'h7',
                'h8', 'br', 'b', 'i', 'strong', 'em', 'a',
                'pre', 'code', 'img', 'tt', 'div', 'ins',
                'del', 'sup', 'sub', 'p', 'ol', 'ul', 'table',
                'thead', 'tbody', 'tfoot', 'blockquote', 'dl',
                'dt', 'dd', 'kbd', 'q', 'samp', 'var', 'hr', 'ruby',
                'rt', 'rp', 'li', 'tr', 'td', 'th', 's', 'strike',
                'summary', 'details', 'caption', 'figure',
                'figcaption', 'abbr', 'bdo', 'cite', 'dfn',
                'mark', 'small', 'span', 'time', 'wbr'}
"""Tags allowed in any HTML input"""

ALLOWED_ATTRS = {
    "*": {'abbr', 'accept', 'accept-charsetaccesskey',
          'action', 'align', 'altaria-describedby',
          'aria-hidden', 'aria-label', 'aria-labelledbyaxis',
          'border', 'cellpadding', 'cellspacing',
          'charcharoff', 'charset', 'checkedclear',
          'cols', 'colspan', 'colorcompact',
          'coords', 'datetime', 'dirdisabled',
          'enctype', 'for', 'frameheaders',
          'height', 'hreflanghspace', 'ismap',
          'label', 'langmaxlength', 'media',
          'methodmultiple', 'name', 'nohref',
          'noshadenowrap', 'open', 'progress',
          'prompt', 'readonly', 'rel', 'revrole',
          'rows', 'rowspan', 'rules', 'scopeselected',
          'shape', 'size', 'spanstart', 'summary',
          'tabindex', 'targettitle', 'type', 'usemap',
          'valign', 'valuevspace', 'width', 'itemprop'},
    "a": {'href'},
    "img": {'src', 'longdesc'},
    "div": {'itemscope', 'itemtype'},
    "blockquote": {'cite'},
    "del": {'cite'},
    "ins": {'cite'},
    "q": {'cite'}
}
"""Attributes allowed in any HTML inputs"""

ALLOWED_PROTO = {'http://*', 'https://*', 'mailto:*'}
"""Procols allowed for URLs in HTML tags"""


def remove_spaces(string: str) -> str:
    """Removes all whitespaces from the given string"""
    if not string:
        return ""
    return "".join(l for l in str(string) if l not in WHITESPACES)


def alphabetic(string: str, decimals: bool = True) -> str:
    """Removes all of the non alphabetical letters from the string"""
    if not string:
        return ""
    if decimals:
        return "".join(l for l in str(string) if l.isalpha() or l.isdecimal())
    return "".join(l for l in str(string) if l.isalpha())


def sanitize_http_method(method: str) -> str:
    """Sanitizes the given HTTP method to normalize it"""
    method = remove_spaces(method).upper()
    if method not in utils.types.HTTPMethod.ACCEPTED and method != "*":
        utils.logging.logger.warn(message="The provided HTTP method {method} does not seem "
                                  "to be in the set of defined HTTP methods".format(method=method))
    return method


def sort_http_methods(methods: typing.Iterable) -> typing.List[str]:
    """Sorts the given HTTP methods to normalize them"""
    methods = {sanitize_http_method(method) for method in methods}
    results = []
    for method in ["GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS"]:  # ordered
        if method in methods:
            results.append(method)
            methods.discard(method)
    results.extend(methods)  # remaining methods

    return results


def sanitize_text(text: str, strict: bool = True) -> str:
    """Sanitize text by removing any forbidden HTML part snippet"""
    if strict:
        # pylint: disable=no-member
        return nh3.clean(str(text), tags={"b", "i", "em", "strong"}, attributes={}, url_schemes=set())
    # pylint: disable=no-member
    return nh3.clean(str(text), tags=ALLOWED_TAGS, attributes=ALLOWED_ATTRS, url_schemes=ALLOWED_PROTO)


def split_on_uppercase(string: str) -> typing.List[str]:
    """
    Splits a string on any uppercase letter

    Parameters
    -----------
    string: str
        The string to split

    Returns
    -------
    list
        The string splitted
    """
    start = 0
    results = []
    string = str(string)
    length = len(string)
    for index, letter in enumerate(string):
        if letter.isupper() and (string[index - 1].islower() or (length > index + 1 and string[index + 1].islower())):
            results.append(string[start:index])
            start = index
    results.append(string[start:])
    return results

DELIMITER = "🈑"


def to_path(name: str) -> str:
    """
    Converts a method name to a path

    Parameters
    ----------
    name: str
        The name of the method/class/object
    """

    name = str(name)

    # hello__username__ --> hello/<username>
    in_variable = False
    result = ""
    name_length = len(name)
    for index, letter in enumerate(name):
        if letter == "_":
            if in_variable:  # <hello_world>
                # the previous _ got replaced by >, so we don't need to convert it too
                if result[-1] == ">":
                    result += "_"
                    continue
                # the previous _ got replaced by <, so we don't need to convert it too
                elif result[-1] == "<":
                    continue
                # "__" --> ">"
                elif index + 1 < name_length and name[index + 1] == "_":
                    in_variable = False
                    result += ">"
                else:
                    result += DELIMITER  # we want to keep any "_" inside a variable name
                    continue
            else:
                # "__" --> "<"
                if index + 1 < name_length and name[index + 1] == "_":
                    in_variable = True
                    result += "_<"
                else:  # a regular _, to be converted into "/"
                    result += letter
                    continue
        else:
            result += letter
    name = result
    # nasse.utils.sanitize --> nasse/utils/sanitize
    name = "/".join(str(name).split("."))
    # hello_world --> hello/world
    name = "/".join(name.split("_"))
    # helloWorld --> hello-world
    name = "-".join(split_on_uppercase(name)).lower()
    return "/" + name.replace(DELIMITER, "_").replace("//", "/").strip("/")

# pylint: disable=invalid-name


def toCamelCase(string: str) -> str:
    """
    Converts a string to camel case

    Parameters
    ----------
        string: str
            The string to convert
    """
    string = str(string)
    if string.isupper():
        return string
    split = string.split("_")  # split by underscore
    final_split = []
    for s in split[(1 if string.startswith("_") else 0):]:
        final_split.extend(s.split(" "))  # split by space
    return ("_" if string.startswith("_") else "") + "".join(l.capitalize() if index > 0 else l for index, l in enumerate(final_split))



"""
File containing the different exceptions which can be raised in Nasse
"""

import typing
from nasse.exceptions.base import NasseException
from nasse.exceptions import arguments, authentication, request, validate


class NasseExceptionTBD(NasseException):
    """
    If an exception is expected to be defined later on
    """



"""
Generates Python usage examples
"""

import json
import typing

from nasse import models


def create_python_example_for_method(endpoint: models.Endpoint, method: str) -> str:
    """
    Generates a Python example for the given endpoint

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to create the example for
    method: str
        The method to create the example for

    Returns
    -------
    str
        The example
    """
    
    params = {param.name: (param.description or param.name)
               for param in models.get_method_variant(method, endpoint.parameters)
               if param.required}
    headers = {header.name: (header.description or header.name)
               for header in models.get_method_variant(method, endpoint.headers)
               if header.required}

    params_render = ""
    headers_render = ""

    if len(params) > 0:
        params_render = ",\n        params = {}".format(json.dumps(params,
                                                                   ensure_ascii=False,
                                                                   indent=4)
                                                        .replace("\n", "\n        "))

    if len(headers) > 0:
        headers_render = ",\n        headers = {}".format(json.dumps(headers,
                                                                     ensure_ascii=False,
                                                                     indent=4)
                                                          .replace("\n", "\n        "))

    return '''import requests
r = requests.request("{method}", "{path}"{params}{headers})
if r.status_code >= 400 or not r.json()["success"]:
    raise ValueError("An error occured while requesting for {path}, error: " + r.json()["error"])
print("Successfully requested for {path}")
print(r.json()["data"])'''.format(method=method,
                                  path=endpoint.path,
                                  params=params_render,
                                  headers=headers_render)


def create_python_example(endpoint: models.Endpoint) -> typing.Dict[str, str]:
    """
    Generates Python examples for the given endpoint

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to create the example for

    Returns
    -------
    dict[str, str]
        A dictionary containing {method: example} values
    """
    results = {}
    for method in endpoint.methods:
        results[method] = create_python_example_for_method(endpoint=endpoint,
                                                           method=method)
    return results



"""Defines custom headers"""
from textual.widgets import Label
from textual.reactive import reactive

from nasse.tui.widget import Widget


class StickyHeader(Widget):
    """A sticky header"""
    DEFAULT_CSS = """
    StickyHeader {
        align: center top;
        dock: top;
        padding: 1;
        height: auto;
    }
    """
    title: reactive[str] = reactive("Header")

    def __init__(self, title: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.title = title

    def compose(self):
        yield Label(f"[underline][bold]{self.title}[/bold][/underline]")
        # yield SectionTitle(self.title)



"""
Creating `curl` examples
"""

import typing

from nasse import models


def create_curl_example_for_method(endpoint: models.Endpoint, method: str) -> str:
    """
    Creates a `curl` example for the given method

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to create the example from
    method: str
        The method to create the example for

    Returns
    -------
    str
        A usage example for `curl`
    """

    def sanitize_backslashes(element: str):
        """
        Turns the backslashes into triple backslashes

        Parameters
        ---------
        element: str
            The element to sanitize

        Returns
        -------
        str
            The sanitized element
        """
        return str(element).replace("\"", "\\\"")

    params = {param.name: param.description for param in models.get_method_variant(method, endpoint.parameters)}
    headers = {header.name: header.description for header in models.get_method_variant(method, endpoint.headers)}

    params_render = ""
    headers_render = ""

    if len(params) > 0:
        # ref: https://everything.curl.dev/http/post/url-encode
        params_render = "\\\n    --data-urlencode " + "\\\n    --data-urlencode ".join(
            ['"' + sanitize_backslashes(param)
             + '=<' + sanitize_backslashes(description)
             + '>"' for param, description in params.items()]
        ) + " "
        if len(headers) <= 0:
            params_render += "\\\n    "

    if len(headers) > 0:
        # ref: https://everything.curl.dev/http/requests/headers
        headers_render = "\\\n    -H " + '\\\n    -H '.join([
            '"' + sanitize_backslashes(header)
            + ': ' + sanitize_backslashes(description)
            + '"' for header, description in headers.items()]
        ) + " "
        headers_render += "\\\n    "

    return '''curl -X {method} {params}{headers}"{path}"'''.format(
        method=method,
        params=params_render,
        headers=headers_render,
        path=endpoint.path
    )


def create_curl_example(endpoint: models.Endpoint) -> typing.Dict[str, str]:
    """
    Creates a `curl` example command to use the endpoint.

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to create the example for

    Returns
    -------
    dict[str, str]
        A dictionary of {method: example} values.
    """
    results = {}
    for method in endpoint.methods:
        results[method] = create_curl_example_for_method(endpoint=endpoint, method=method)
    return results



"""Displays a series nicely"""

import statistics
import typing

from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widgets import Label, Sparkline

from nasse.tui.widget import Widget
from nasse.localization import EnglishLocalization, Localization


def transform_time(value: typing.Union[int, float]):
    """
    Transforms the milliseconds to a human readable but also width fixed string

    Here are the different transformations :
    5 slots available: xxxxx
    | Unit             | First | Second| Third | Fourth|
    | ---------------- | ----- | ----- | ----- | ----- |
    | 1 d = 24 h       | 9d23h | 99  d | 999 d | 9999d |
    | 1 h = 3600 s     | 9h59m | 99h59 |       |       |
    | 1 s = 1000 ms    | 9.99s | 99.9s | 999 s | 9999s |
    | 1 ms             | 9.9ms | 99 ms | 999ms |       |
    """
    if value >= 360_000_000:  # 100 hours
        # days format
        value /= 86_400_000  # 1 day
        days = int(value)
        value -= days
        hours = int(value * 24)  # 1 hour
        if days >= 10:
            return f"{days}{' ' * (4 - len(str(days)))}d"
        else:
            return f"{days}d{str(hours).zfill(2)}h"
    elif value >= 10_000_000:  # 10 000 seconds
        # hours format
        value /= 3_600_000
        hours = int(value)
        value -= hours
        minutes = int(value * 60)  # 1 minute
        if hours >= 10:
            return f"{hours}h{str(minutes).zfill(2)}"
        else:
            return f"{hours}h{str(minutes).zfill(2)}m"
    elif value >= 1000:  # 1 000 milliseconds
        # seconds format
        seconds = value / 1000
        if seconds >= 1000:
            return f"{int(seconds)}s"
        elif seconds >= 100:
            return f"{int(seconds)} s"
        else:
            digits = 1 if seconds >= 10 else 2
            seconds = str(round(seconds, digits))
            while len(seconds) < 4:
                seconds += "0"
            return f"{seconds}s"
    else:
        # milliseconds format
        if value >= 100:
            return f"{int(value)}ms"
        elif value >= 10:
            return f"{int(value)} ms"
        else:
            value = str(float(round(value, 1)))
            while len(value) < 3:
                value += " "
            return f"{value}ms"


class TimeSeries(Widget):
    """A time series widget"""
    DEFAULT_CSS = """
    TimeSeries {
        height: auto;
        width: 34;
    }
    TimeSeries > Label {
        margin: 0 1 0 1;
        align-horizontal: center;
    }
    """
    series: reactive[typing.List[int]] = reactive(list)

    def __init__(self, series: typing.Optional[typing.List[int]] = None, localization: typing.Type[Localization] = EnglishLocalization, **kwargs) -> None:
        super().__init__(**kwargs)
        self.series = series or []
        self.localization = localization

    @property
    def statistics(self):
        """Returns a statistics banner for the series"""
        min_ping = min(self.series) if self.series else 0
        max_ping = max(self.series) if self.series else 0
        avg_ping = statistics.mean(self.series) if self.series else 0

        return f"{self.localization.tui_min} {transform_time(min_ping)} / {self.localization.tui_average} {transform_time(avg_ping)} / {self.localization.tui_max} {transform_time(max_ping)}"

    def compose(self) -> ComposeResult:
        yield Label(self.statistics)
        yield Sparkline(
            self.series,
            summary_function=statistics.mean
        )

    def watch_series(self, series: typing.List[int]):
        """When the `series` attribute is set"""
        try:
            self.query_one(Label).update(self.statistics)
            self.query_one(Sparkline).data = series
        except Exception as err:
            print("WARNING: Couldn't update view;", err)



"""
The french translation

Copyright
---------
Animenosekai
    Original Author, MIT License
"""

from nasse.localization.base import Localization


class FrenchLocalization(Localization):
    """
    The french translation for the docs generation
    """
    __native__ = "Français"

    sections = "Sections"
    getting_started = "Pour commencer"

    yes = "Oui"
    no = "Non"

    no_description = "Il n'y a pas de description"
    using_method = "En utilisant {method}"

    # Authentication
    authentication = "Authentification"
    no_auth_rule = "Il n'y a pas de règle d'authentification définie"
    no_login = "Il n'est **pas** nécessaire d'être authentifié"
    login_with_types_required = "Il est **nécessaire** d'être authentifié avec un compte {types}"
    login_with_types_required = "Il est **optionnel** d'être authentifié avec un compte {types}"
    login_required = "Il est **nécessaire** d'être authentifié"
    login_optional = "Il est **optionnel** d'être authentifié"
    login_suffix_only_verified = " (mais l'authentification est seulement vérifié et le compte ne sera pas cherché)"

    # User sent values
    parameters = "Paramètres"
    headers = "En-têtes"
    cookies = "Cookies"
    dynamic_url = "URL Dynamique"

    name = "Nom"
    description = "Description"
    required = "Obligatoire"
    type = "Type"

    # Example
    example = "Exemple"

    # Response
    response = "Réponse"
    example_response = "Exemple de réponse"
    not_json_response = "Cet endpoint ne semble pas retourner de réponse formatté en JSON."

    # Response description
    returns = "Retourne"
    field = "Champ"
    nullable = "Peut être `null`"

    # Errors
    possible_errors = "Erreurs possibles"
    exception = "Erreur"
    code = "Code"

    # Index
    index = "Index"
    return_to_index = "Retourner à l'Index"

    # Postman
    postman_description = "Tous les endpoints sous la section '{section}' de l'interface API de {name}"

    # Headers
    getting_started_header = '''
# Référence d'API pour {name}

Bienvenue sur la référence d'API pour {name}.

## Globalité

### Format de réponse

Généralement, les réponses JSON seront formattez comme suit (même lorsque des erreurs critiques sont encontrées)

```json
{{
    "success": true,
    "message": "Ça marche!",
    "error": null,
    "data": {{}}
}}
```

| Champ        | Description                                      | Peut être `null` |
| ------------ | ------------------------------------------------ | ---------------- |
| `success`    | Si la requête est un succès ou pas               | Non            |
| `message`    | Un message qui décrit ce qu'il s'est passé       | Oui             |
| `error`      | Le nom de l'erreur si il y en a une              | Oui             |
| `data`       | Des données supplémentaires, les informations demandées | Non            |

### Erreurs

De multiples erreurs peuvent survenir, que ce soit du côté client (au niveau de la requête) ou du côté du serveur.

Les erreurs spécifiques sont documentés dans chaque *Endpoint* mais celles-ci sont celles qui peuvent survenir sur n'importe quel endpoint :

| Erreur                      | Description                                                                                                     | Code  |
| --------------------------- | --------------------------------------------------------------------------------------------------------------- | ----- |
| `SERVER_ERROR`              | Quand une erreur survient sur {name} en traitant la requête.                                                    | 500   |
| `MISSING_CONTEXT`           | Quand dans le code vous essayez d'accéder à une ressource seulement disponible dans un contexte Nasse (pendant une requête) mais que vous n'êtes pas dans un contexte Nasse.       | 500   |
| `INTERNAL_SERVER_ERROR`     | Quand une erreur critique survient sur le système                                                               | 500   |
| `METHOD_NOT_ALLOWED`        | Quand vous faites une requête avec le mauvaise méthode HTTP                                                     | 405   |
| `CLIENT_ERROR`              | Quand quelque chose manque où n'est pas bon avec la requête                                                     | 400   |
| `INVALID_TYPE`              | Quand Nasse n'a pas pu convertir la valeur qui a été envoyée dans le bon type                                   | 400   |
| `MISSING_VALUE`             | Quand quelque chose manque à la requête                                                                         | 400   |
| `MISSING_PARAM`             | Quand un paramètre manque à la requête                                                                          | 400   |
| `MISSING_DYNAMIC`           | Quand une valeur dynamique de l'URL manque                                                                      | 400   |
| `MISSING_HEADER`            | Quand un en-tête manque à la requête                                                                            | 400   |
| `MISSING_COOKIE`            | Quand un cookie manque à la requête                                                                             | 400   |
| `AUTH_ERROR`                | Quand une erreur d'authentification survient                                                                    | 403   |

### Requêtes authentifiées

Quand un utilisateur doit être connecter, l'en-tête "Authorization" doit être définie avec le token de connexion communiqué lorsqu'il s'est connecté.

Vous pouvez aussi utiliser le paramètre "{id}_token" ou le cookie "__{id}_token" mais ceux-ci ne vont pas être priorisé.

Si la règle d'authentification de l'*endpoint* est défini comme "seulement vérifié", le compte ne doit pas être cherché dans la base de donnée mais la forme du token ou des informations comprises à l'intérieur peuvent être vérifiées.

### Mode de développement

Lorsque le mode de développement est activé (`-d` ou `--debug`), de multiples informations sont renvoyées dans la section `debug` des réponses et le niveau `DEBUG` de logging est sélectioné sur le serveurr.

La section 'debug' est disponible sur tous type d'erreur, sauf celles qui sont renvoyées par Flask de manière interne (comme `INTERNAL_SERVER_ERROR`, `METHOD_NOT_ALLOWED`, etc.). En effet, celles-ci doivent modifier le moins possible la requête pour ne pas se retrouver avec une autre erreur qui pourrait survenir)

Le champ "call_stack" est activé seulement quand il y a le paramètre `call_stack` qui est envoyé avec la requête.

```json
{{
    "success": true,
    "message": "Nous n'avons pas pu satisfaire la requête",
    "error": null,
    "data": {{
        "username": "Animenosekai"
    }},
    "debug": {{
        "time": {{
            "global": 0.036757,
            "verification": 0.033558,
            "authentication": 0.003031,
            "processing": 4.9e-05,
            "formatting": 0.0001
        }},
        "ip": "127.0.0.1",
        "headers": {{
            "Host": "api.{id}.com",
            "Connection": "close",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            "Accept-Language": "fr-fr",
            "Accept-Encoding": "gzip, deflate, br",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
        }},
        "values": {{}},
        "domain": "api.{id}.com",
        "logs": [
            "1636562693.036563｜[INFO] [nasse.receive.Receive.__call__] → Incoming GET request to /account/name from 127.0.0.1",
            "1636562693.070008｜[ERROR] [nasse.exceptions.base.MissingToken.__init__] An authentication token is missing from the request"
        ],
        "call_stack": [
            "pass the 'call_stack' parameter to get the call stack"
        ]
    }}
}}
```

'''

    section_header = '''
# Référence de la section {name}

Ce fichier liste et explique les différents *endpoints* disponible sous la section {name}
'''

    # TUI

    # Footer
    tui_history = "Historique"
    tui_result = "Résultat"
    tui_explorer = "Explorateur"
    tui_submit = "Envoyer"
    tui_options = "Options"
    tui_quit = "Quitter"

    # History
    # WARNING: This should stay under 3 characters to avoid having styling issues
    tui_min = "Min"
    tui_average = "Moy"
    tui_max = "Max"

    # Explorer
    tui_reset = "Réinitialiser"

    # Request
    tui_request = "Requête"
    tui_name = "nom"
    tui_value = "valeur"
    tui_path = "chemin"
    # tui_parameters = "Parameters"
    # tui_headers = "Headers"
    # tui_cookies = "Cookies"
    tui_file = "Fichier"
    tui_add_file = "Ajouter un fichier"
    tui_data = "Donnée"
    tui_add_data_file = "Ajouter un fichier de données"

    # Options
    tui_language = "Langue"
    tui_language_notice = "Vous devez relancer l'application pour que les changements prennent effet"
    tui_base_url = "URL de base"
    tui_base_url_placeholder = "L'URL de base pour les requêtes et l'explorateur"
    tui_endpoints_update = "Mise à jour des endpoints"
    tui_endpoints_update_placeholder = "Fréquence de mise à jour des endpoints pour l'explorateur (sec.)"
    tui_history_limit = "Limite d'historique"
    tui_history_limit_placeholder = "Le nombre maximale de requêtes dans l'historique"
    tui_timeout = "Limite de temps"
    tui_timeout_placeholder = "Limite de temps pour le traitement d'une requête (sec.)"
    tui_redirects = "Redirection"
    tui_allow_redirects = "Autoriser les redirections"
    tui_proxies = "Proxies"
    tui_security = "Sécurité"
    tui_verify_request = "Vérification des requêtes"
    tui_certificate_files = "Certificats"
    tui_add_certificate = "Ajouter un certificat"

    # Result
    tui_start_prompt = "Commencez par envoyer une requête"
    tui_content = "Contenu"
    tui_no_content = "Impossible d'afficher le contenu"
    tui_contacting = "Appel de {url}"
    tui_files = "Fichiers"
    tui_error = "Erreur"

    # File Explorer
    tui_filter = "Filtre"

    # Quit
    tui_quit_confirmation = "Voulez-vous vraiment quitter ?"
    tui_cancel = "Revenir"



"""
Nasse configuration
"""

import dataclasses
import pathlib
import typing
import urllib.parse
from nasse.utils import formatter
from nasse import __info__

def _alphabetic(string):
    return "".join(l for l in str(string) if l.isalpha() or l.isdecimal())


@dataclasses.dataclass
class NasseConfig:
    """
    A configuration object for a Nasse app
    """

    def verify_logger(self):
        """Verifies the logger"""
        from nasse.utils.logging import Logger
        if self.logger is None:
            self.logger = Logger(self)
            # self.logger = Logger()

    def verify_logging_level(self):
        """Verifies the given logging level"""
        from nasse.utils.logging import LoggingLevel

        if isinstance(self.logging_level, LoggingLevel):
            level_name = self.logging_level.name
        else:
            level_name = str(self.logging_level)

        level_name = level_name.upper().replace(" ", "")

        if not self.logging_level:
            self.logging_level = LoggingLevel.DEBUG if self.debug else LoggingLevel.INFO
        elif level_name == "ERROR":
            self.logging_level = LoggingLevel.ERROR
        elif level_name == "WARNING":
            self.logging_level = LoggingLevel.WARNING
        elif level_name == "INFO":
            self.logging_level = LoggingLevel.INFO
        elif level_name == "DEBUG":
            self.logging_level = LoggingLevel.DEBUG
        elif level_name == "HIDDEN":
            self.logging_level = LoggingLevel.HIDDEN
        else:
            try:
                self.logger.warn(f"Couldn't understand the logging level {self.logging_level}. Defaulting to `INFO`")
            except Exception:
                pass
            self.logging_level = LoggingLevel.INFO

    def __setattr__(self, __name: str, __value: typing.Any) -> None:
        if __name == "debug" and (not isinstance(self.logging_level, str)) and (not self.logging_level or self.logging_level.value < 4):
            from nasse.utils.logging import LoggingLevel
            if __value:
                self.logging_level = LoggingLevel.DEBUG
            else:
                self.logging_level = LoggingLevel.INFO
        super().__setattr__(__name, __value)

    def __post_init__(self):
        # self.VERSION = __info__.__version__

        if not self.id:
            self.id = _alphabetic(self.name).lower()

        self.verify_logger()
        self.verify_logging_level()

        if not self.log_file:
            self.log_file = (pathlib.Path() / ".nasse" / "debug" / "log") if self.debug else None

        if isinstance(self.cors, str):
            rule = str(self.cors).replace(" ", "")
            if rule == "*":
                self.cors = ["*"]
            else:
                parsed = urllib.parse.urlparse(rule)
                netloc = (parsed.netloc if parsed.netloc
                          else parsed.path.split("/")[0])
                scheme = parsed.scheme or "https"
                rule = f'{scheme}://{netloc}'
                self.cors = [rule]
        elif isinstance(self.cors, bool):
            self.cors = ["*"] if self.cors else []
        else:
            self.cors = []
            for rule in self.cors:
                rule = str(rule).replace(" ", "")
                if rule == "*":
                    self.cors.append("*")
                    continue
                else:
                    parsed = urllib.parse.urlparse(rule)
                    netloc = parsed.netloc if parsed.netloc else parsed.path.split("/")[0]
                    scheme = parsed.scheme or "https"
                    rule = f'{scheme}://{netloc}'
                    self.cors.append(rule)

        self.server_header = formatter.format(self.server_header, config=self)

    name: str = "Nasse"
    id: typing.Optional[str] = None # pylint: disable=invalid-name
    host: str = "127.0.0.1"
    port: int = 5005
    debug: bool = False
    account_management: typing.Optional["AccountManagement"] = None
    cors: typing.Union[str, bool, typing.Iterable] = True
    max_request_size: int = int(1e+9)
    compress: bool = True
    log_file: typing.Optional[pathlib.Path] = None
    logging_level: typing.Optional["LoggingLevel"] = "INFO"
    logger: typing.Optional["Logger"] = None
    server_header: str = "nasse/{version} ({name})"
    sanitize_user_input: bool = True
    base_dir: pathlib.Path = pathlib.Path().resolve().absolute()



"""
Nasse logging utilities

Copyright
---------
Animenosekai
    Original author, MIT License
"""
import dataclasses
import datetime
import enum
import linecache
import typing

import rich.console
from nasse.utils import formatter


class LoggingLevel(enum.Enum):
    """
    A logging level
    """
    ERROR = 1
    WARNING = 2
    INFO = 3
    DEBUG = 4
    HIDDEN = 999


@dataclasses.dataclass
class Record:
    """
    Represents a record in the log stack
    """

    def __post_init__(self):
        self.time = datetime.datetime.now()

    level: LoggingLevel
    msg: str


class Logger:
    """
    A Nasse logging object, the logger used thoughout your app
    """

    TIME_FORMAT: typing.Union[str, typing.Callable[[datetime.datetime], typing.Any]] = "%Y/%m/%d, %H:%M:%S"
    # TIME_FORMAT = lambda time: int(time.timestamp())
    """
    The logging time format

    It can be either a function taking a datetime.datetime object or a string,
    which will be passed to datetime.datetime.strftime
    """

    TEMPLATES = {
        LoggingLevel.INFO: "{grey}{time} | {normal}[{level}] ({app}) {message}",
        LoggingLevel.DEBUG: "{grey}{time} | [{level}] ({app}) {message}{normal}",
        LoggingLevel.WARNING: "{grey}{time} |{normal} [{level}] ({app}) {yellow}{message}{normal}",
        LoggingLevel.ERROR: "{grey}{time} |{normal} [{level}] ({app}) {red}{message}{normal}"
    }

    def __init__(self, config: "config.NasseConfig" = None) -> None:
        self.config = config
        if not self.config:
            from nasse.config import NasseConfig

            class NewConfig(NasseConfig):
                def verify_logger(self):
                    return

            self.config = NewConfig()
        self.recording = False
        self.record = []

        self._rich_console = rich.console.Console()

        if self.config.log_file:
            self.config.log_file.parent.mkdir(parents=True, exist_ok=True)
            self.config.log_file.touch(exist_ok=True)
            with open(self.config.log_file, "a") as f:
                WIDTH = 32
                MESSAGE = "LOG START"
                padding = WIDTH - len(MESSAGE) // 2
                f.write(("=" * padding) + MESSAGE + ("=" * padding) + "\n")

    def log(self, *msg,
            level: LoggingLevel = LoggingLevel.INFO,
            end: str = "\n",
            sep: str = " ", **kwargs) -> None:
        """
        Logging the given message to the console.
        """
        if level.value > self.config.logging_level.value:
            return

        result = formatter.format(
            str(sep).join(str(m) for m in msg),
            time_format=self.TIME_FORMAT,
            config=self.config,
            level=level.name,
            **kwargs
        )

        record_output = result
        for element in formatter.Colors:
            # removing the colors for any file or recording output
            record_output = record_output.replace(element.value, "")

        if self.recording:
            self.record.append(Record(level=level, msg=record_output))

        if self.config.log_file:
            self.write_to_file(msg=record_output, level=level, **kwargs)

        template = self.TEMPLATES.get(level, "{message}")
        result = formatter.format(template, time_format=self.TIME_FORMAT, config=self.config, level=level.name, message=result, **kwargs)

        print(result, end=str(end))

    __call__ = log

    def write_to_file(self,
                      msg: str,
                      level: LoggingLevel = LoggingLevel.INFO,
                      **kwargs) -> None:
        """
        Internal function called to write to the log file
        """
        if not self.config.log_file:
            return

        self.config.log_file.parent.mkdir(parents=True, exist_ok=True)
        with open(self.config.log_file, "a", encoding="utf-8") as f:
            f.write(formatter.format("[{level}] ({name}) - {time} - {msg}\n",
                                     time_format=lambda time: int(time.timestamp()),
                                     level=level.name,
                                     name=self.config.name,
                                     msg=msg,
                                     **kwargs))

    def info(self, *msg, **kwargs):
        """
        Logs the given message with the `INFO` level
        """
        self.log(*msg, level=LoggingLevel.INFO, **kwargs)

    # aliasing `info`
    print = info

    def debug(self, *msg, **kwargs) -> None:
        """
        Logs the given message with the `DEBUG` level
        """
        self.log(*msg, level=LoggingLevel.DEBUG, **kwargs)

    def warning(self, *msg, **kwargs) -> None:
        """
        Logs the given message with the `WARNING` level
        """
        self.log(*msg, level=LoggingLevel.WARNING, **kwargs)

    warn = warning

    def error(self, *msg, **kwargs) -> None:
        """
        Logs the given message with the `ERROR` level
        """
        self.log(*msg, level=LoggingLevel.ERROR, **kwargs)

    def hidden(self, *msg, **kwargs) -> None:
        """
        Logs the given message with the `HIDDEN` level

        It only writes to the log file and the records
        """
        self.log(*msg, level=LoggingLevel.HIDDEN, **kwargs)

    hide = hidden

    def __enter__(self):
        """
        Begins recording the output
        """
        self.recording = True
        self.record = []
        return self

    def __exit__(self, type, value, traceback):
        """
        Stops recording the output
        """
        self.recording = False

    def print_exception(self,
                        show_locals: bool = False,
                        **kwargs) -> None:
        """
        Prints the latest exception, nicely

        Parameters
        ----------
        show_locals: bool, default = False
            When enabled, shows the local variables to the console
        """
        self._rich_console.print_exception(show_locals=show_locals, **kwargs)

    exception = print_exception


RECORDING = False
CALL_STACK = []


class StackFrame:
    """
    A call stack frame
    """

    def __init__(self, frame) -> None:
        # print(dir(frame))
        # print(dir(frame.f_code))
        self.name = frame.f_code.co_name
        self.filename = frame.f_code.co_filename
        self.lineno = frame.f_lineno
        self.back_frame = frame.f_back
        self._line = None
        self._calling_line = None

    def __repr__(self) -> str:
        return "<NasseStackFrame '{name}' {filename} on {line_number}>".format(name=self.name, filename=self.filename, line_number=self.lineno)

    @property
    def line(self):
        if self._line is None:
            self._line = linecache.getline(self.filename, self.lineno)
        return self._line.strip()

    @property
    def calling_line(self):
        if self._calling_line is None:
            self._calling_line = linecache.getline(
                self.back_frame.f_code.co_filename, self.back_frame.f_lineno)
        return self._calling_line.strip()

    def as_dict(self) -> dict:
        return {
            "name": self.name,
            "filename": self.filename,
            "lineNumber": self.lineno,
            "calledBy": {
                "name": self.back_frame.f_code.co_name,
                "filename": self.back_frame.f_code.co_filename,
                "lineNumber": self.back_frame.f_code.co_firstlineno
            }
        }


class CallStackRecorder:
    """
    A call stack recorder
    """

    def __enter__(self):
        """
        Begins recording the calls
        """
        global RECORDING
        global CALL_STACK
        RECORDING = True
        CALL_STACK = []
        return self

    @property
    def recording(self):
        return RECORDING

    @property
    def call_stack(self):
        return CALL_STACK

    def __exit__(self, type, value, traceback):
        """
        Stops recording
        """
        global RECORDING
        RECORDING = False


def _generate_trace(config):
    """
    Internal function to generate a trace to record the call stack
    """
    def add_to_call_stack(frame, event, arg):
        """
        Internal function to add a call to the call stack
        """
        if RECORDING and event == "call" and frame.f_code.co_filename.startswith(str(config.base_dir)):
            CALL_STACK.append(StackFrame(frame))
        return None
    return add_to_call_stack


logger = Logger()


def log(*msg,
        level: LoggingLevel = LoggingLevel.INFO,
        end: str = "\n",
        sep: str = " ", **kwargs) -> None:
    """
    Logging the given message to the console.
    """
    logger.log(*msg, level=level, end=end, sep=sep, **kwargs)


info = log


def debug(*msg, **kwargs) -> None:
    """
    Logs the given message with the `DEBUG` level
    """
    logger.debug(*msg, **kwargs)


def warning(*msg, **kwargs) -> None:
    """
    Logs the given message with the `WARNING` level
    """
    logger.warning(*msg, **kwargs)


warn = warning


def error(*msg, **kwargs) -> None:
    """
    Logs the given message with the `ERROR` level
    """
    logger.error(*msg, **kwargs)


def hidden(*msg, **kwargs) -> None:
    """
    Logs the given message with the `HIDDEN` level

    It only writes to the log file and the records
    """
    logger.hidden(*msg, **kwargs)


hide = hidden



"""A web server framework written on top of Flask which lets you focus on your ideas 🍡"""

# autopep8: off
from .__info__ import __version__, __license__, __author__, __copyright__ # isort:skip
from . import docs # isort:skip
from flask import g # isort:skip
from flask.wrappers import Request as FlaskRequest # isort:skip
from flask.wrappers import Response as FlaskResponse # isort:skip

from .nasse import Nasse # isort:skip
from .config import NasseConfig # isort:skip
from .request import Request # isort:skip
from .response import Response # isort:skip
from .models import * # isort:skip
from . import logging # isort:skip
from .utils.logging import Logger, LoggingLevel, logger, log, debug # isort:skip


class RequestProxy(FlaskRequest, Request):
    """Proxies the global `Request` object"""
    def __init__(self) -> None:
        return

    def __getattribute__(self, name: str):
        return getattr(g.request, name)


request = RequestProxy()

# For backward compatibility
# old_name = new_name
# if something has been renamed in the new versions

# autopep8: on


"""
Exceptions handling the authentication errors
"""

import typing
from nasse.exceptions.base import NasseException


class AuthenticationError(NasseException):
    """When an error occurs with the user authentication step"""
    MESSAGE = "An error occured while checking your authentication"
    EXCEPTION_NAME = "AUTH_ERROR"
    STATUS_CODE = 403


class MissingToken(AuthenticationError):
    """When the token is missing to authenticate"""
    LOG = False
    MESSAGE = "The authentication token is missing from your request"


class Forbidden(AuthenticationError):
    """When the user is not allowed to access something"""
    MESSAGE = "You are not allowed to access this endpoint"



"""
A pretty accurate timer to measure code
"""

import time
import sys

if sys.version_info > (3, 7):
    ns_clock = time.perf_counter_ns  # novermin
else:
    def ns_clock():
        """Performance clock counting in nanoseconds"""
        return int(time.perf_counter() * 1e+9)


class Timer():
    """
    A pretty accurate timer
    """

    def __init__(self) -> None:
        """
        A timer to measure performance of a piece of code

        Example
        -------
        >>> from nasse.timer import Timer
        >>> with Timer() as timer:
        ...     # do heavy stuff
        >>> timer.time
        5.628793271 # in seconds
        >>> timer.time_ns
        5628793271 # in nanoseconds
        """
        self.current_timer = ns_clock()
        self.time_ns = 0
        self.time = 0

    def start(self):
        """
        Manually starts the timer
        """
        self.current_timer = ns_clock()
        self.time_ns = 0
        self.time = 0

    def stop(self):
        """
        Save the current time

        Returns
        --------
            float
                The time taken between Timer.start() and now (in seconds)
        """
        self.time_ns = ns_clock() - self.current_timer
        self.time = self.time_ns / 1e+9
        return self.time

    def __enter__(self):
        self.current_timer = ns_clock()
        self.time_ns = 0
        self.time = 0
        return self

    def __exit__(self, type, value, traceback):
        self.time_ns = ns_clock() - self.current_timer
        self.time = self.time_ns / 1e+9



"""Text components"""

from textual.widgets import Label
from nasse.tui.widget import Widget


class SectionTitle(Widget):
    """A section title"""

    DEFAULT_CSS = """
    SectionTitle {
        height: auto;
        margin-top: 2;
        margin-bottom: 1;
    }
    """

    def __init__(self, title: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.title = str(title)

    def compose(self):
        yield Label(f"[underline][bold]{self.title}[/bold][/underline]")



"""Deals with history requests response"""
import typing
from urllib.parse import urlparse

import requests
from textual.reactive import reactive
from textual.widgets import Button, Label

from nasse.tui.components import series


class HistoryResponse(Button):
    """A history response"""

    DEFAULT_CSS = """
    HistoryResponse {
        width: 100%;
        height: 3;
        color: $text;

        border: round white;
        padding: 0 1 0 1;
        height: auto;
        margin: 0 1 0 1;
        opacity: 1;
    }

    HistoryResponse:hover {
        opacity: 0.75;
    }

    Button:focus {}

    Button:hover {}

    .history-response-path {
        margin-bottom: 1;
    }

    .history-response-status {
        opacity: 0.75;
    }
    """

    response: reactive[typing.Union[requests.Response, "Error"]]

    def __init__(self, response: typing.Union[requests.Response, "Error"], **kwargs) -> None:
        super().__init__(**kwargs)
        self.response = response

    def compose(self):
        from nasse.tui.apps.http_app import Error

        url = urlparse(self.response.url)
        yield Label(f"[bold]{url.path}[/bold]", classes="history-response-path")

        if isinstance(self.response, Error):
            message = f"{self.response.method}・[yellow]{self.response.exception.__class__.__name__}[/yellow]"
        elif hasattr(self.response, "ok") and not self.response.ok:
            message = f"{self.response.request.method}・[red]{self.response.status_code}[/red] ({series.transform_time(self.response.elapsed.total_seconds() * 1000)})"
        elif hasattr(self.response, "request") and hasattr(self.response, "status_code") and hasattr(self.response, "elapsed"):
            message = f"{self.response.request.method}・{self.response.status_code} ({series.transform_time(self.response.elapsed.total_seconds() * 1000)})"
        else:
            message = "[yellow]ERROR[/yellow]"

        yield Label(message, classes="history-response-status")

    def render(self):
        return ""

    def _start_active_affect(self):
        return



"""Stores information on the current module version"""

# Authors
__author__ = 'Anime no Sekai'
__maintainer__ = 'Anime no Sekai'
__credits__ = ['animenosekai']
__email__ = 'animenosekai.mail@gmail.com'
__repository__ = "https://github.com/Animenosekai/nasse"

# Module
__module__ = "Nasse"
__status__ = 'Beta'
__year__ = 2023
__license__ = 'MIT License'

__copyright__ = f'Copyright {__year__}, {__module__}'

# PEP 440 Compilant
__version__ = "2.0.2"



"""
XML Conversion

Source: dict2xml by delfick \n
GitHub: https://github.com/delfick/python-dict2xml \n
PyPI: https://pypi.org/project/dict2xml/ \n
Based on version 1.7.0 \n
Licensed under the MIT License

The MIT License (MIT)

Copyright (c) 2018 Stephen Moore

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

-------------------

Arranged for Nasse

It supports bytes, file-like objects, null (None), booleans, and fallback for non-supported types.
The API is also a little bit simpler.
"""
import base64
import collections
import collections.abc
import re
import typing

from nasse import utils

start_ranges = "|".join(
    "[{0}]".format(r)
    for r in [
        "\xC0-\xD6",
        "\xD8-\xF6",
        "\xF8-\u02FF",
        "\u0370-\u037D",
        "\u037F-\u1FFF",
        "\u200C-\u200D",
        "\u2070-\u218F",
        "\u2C00-\u2FEF",
        "\u3001-\uD7FF",
        "\uF900-\uFDCF",
        "\uFDF0-\uFFFD",
    ]
)

NameStartChar = re.compile(r"(:|[A-Z]|_|[a-z]|{0})".format(start_ranges))
NameChar = re.compile(r"(\-|\.|[0-9]|\xB7|[\u0300-\u036F]|[\u203F-\u2040])")

########################
# NODE
########################


class Node(object):
    """
        Represents each tag in the tree

        Each node has _either_ a single value or one or more children
        If it has a value:
            The serialized result is <%(tag)s>%(value)s</%(tag)s>

        If it has children:
            The serialized result is
                <%(wrap)s>
                    %(children)s
                </%(wrap)s>

        Which one it is depends on the implementation of self.convert
    """

    # A mapping of characters to treat as escapable entities and their replacements
    entities = [("&", "&amp;"), ("<", "&lt;"), (">", "&gt;")]

    def __init__(self, wrap="", tag="", data=None, iterables_repeat_wrap=True):
        self.tag = self.sanitize_element(tag)
        self.wrap = self.sanitize_element(wrap)
        self.data = data
        self.type = self.determine_type()
        self.iterables_repeat_wrap = iterables_repeat_wrap

        if self.type == "flat" and isinstance(self.data, str):
            # Make sure we deal with entities
            for entity, replacement in self.entities:
                self.data = self.data.replace(entity, replacement)

    def serialize(self, indenter):
        """Returns the Node serialized as an xml string"""
        # Determine the start and end of this node
        wrap = self.wrap
        end, start = "", ""
        if wrap:
            end = "</{0}>".format(wrap)
            start = "<{0}>".format(wrap)

        # Convert the data attached in this node into a value and children
        value, children = self.convert()

        # Determine the content of the node (essentially the children as a string value)
        content = ""
        if children:
            if self.type != "iterable":
                # Non-iterable wraps all it's children in the same tag
                content = indenter((c.serialize(indenter)
                                   for c in children), wrap)
            else:
                if self.iterables_repeat_wrap:
                    # Iterables repeat the wrap for each child
                    result = []
                    for c in children:
                        content = c.serialize(indenter)
                        if c.type == "flat":
                            # Child with value, it already is surrounded by the tag
                            result.append(content)
                        else:
                            # Child with children of it's own, they need to be wrapped by start and end
                            content = indenter([content], True)
                            result.append("".join((start, content, end)))

                    # We already have what we want, return the indented result
                    return indenter(result, False)
                else:
                    result = []
                    for c in children:
                        result.append(c.serialize(indenter))
                    return "".join([start, indenter(result, True), end])

        # If here, either:
        #  * Have a value
        #  * Or this node is not an iterable
        return "".join((start, value, content, end))

    def determine_type(self):
        """
            Return the type of the data on this node as an identifying string

            * Iterable : Supports "for item in data"
            * Mapping : Supports "for key in data: value = data[key]"
            * flat : A string or something that isn't iterable or a mapping
        """
        data = self.data
        if isinstance(data, (int, float, str)):
            return "flat"
        elif isinstance(data, bool):
            return "bool"
        elif data is None:
            return "null"
        elif utils.unpack.is_unpackable(data):
            return "mapping"
        elif isinstance(data, bytes):
            return "bytes"
        elif hasattr(data, "read") and hasattr(data, "tell") and hasattr(data, "seek"):
            return "file"
        elif isinstance(data, typing.Iterable):
            return "iterable"
        else:
            utils.logging.logger.debug("Object of type <{type}> will be converted to str while encoding to XML".format(type=data.__class__.__name__))
            return "flat"

    def convert(self):
        """
            Convert data on this node into a (value, children) tuple depending on the type of the data
            If the type is :
                * flat : Use self.tag to surround the value. <tag>value</tag>
                * mapping : Return a list of tags where the key for each child is the wrap for that node
                * iterable : Return a list of Nodes where self.wrap is the tag for that node
        """
        val = ""
        typ = self.type
        data = self.data
        children = []

        if typ == "mapping":
            sorted_data = data
            if not isinstance(data, collections.OrderedDict):
                sorted_data = sorted(data)

            for key in (sorted_data or []):
                item = data[key]
                children.append(
                    Node(key, "", item,
                         iterables_repeat_wrap=self.iterables_repeat_wrap)
                )

        elif typ == "iterable":
            for item in data:
                children.append(
                    Node("", self.wrap, item,
                         iterables_repeat_wrap=self.iterables_repeat_wrap,)
                )
        elif typ == "bool":
            val = "true" if data else "false"
            if self.tag:
                val = "<{0}>{1}</{2}>".format(self.tag, val, self.tag)
        elif typ == "null":
            val = "null"
            if self.tag:
                val = "<{0}>{1}</{2}>".format(self.tag, val, self.tag)
        elif typ == "bytes":
            val = base64.b64encode(data).decode("utf-8")
            if self.tag:
                val = "<{0}>{1}</{2}>".format(self.tag, val, self.tag)
        elif typ == "file":
            position = data.tell()  # storing the current position
            content = data.read()  # read it (place the cursor at the end)
            data.seek(position)  # go back to the original position
            if "b" in data.mode:  # if binary mode
                val = base64.b64encode(data).decode("utf-8")
            else:
                val = str(content)

            if self.tag:
                val = "<{0}>{1}</{2}>".format(self.tag, val, self.tag)
        else:
            val = str(data)
            if self.tag:
                val = "<{0}>{1}</{2}>".format(self.tag, val, self.tag)

        return val, children

    @staticmethod
    def sanitize_element(wrap):
        """
            Convert `wrap` into a valid tag name applying the XML Naming Rules.

                * Names can contain letters, numbers, and other characters
                * Names cannot start with a number or punctuation character
                * Names cannot start with the letters xml (or XML, or Xml, etc)
                * Names cannot contain spaces
                * Any name can be used, no words are reserved.

            :ref: http://www.w3.org/TR/REC-xml/#NT-NameChar
        """
        if wrap and isinstance(wrap, str):
            if wrap.lower().startswith("xml"):
                wrap = "_" + wrap
            return "".join(
                ["_" if not NameStartChar.match(wrap) else ""]
                + ["_" if not (NameStartChar.match(c)
                               or NameChar.match(c)) else c for c in wrap]
            )
        else:
            return wrap


def encode(data, minify: bool = False):
    """
    Encodes Python data to XML

    Parameters
    ----------
        data: Any
            The data to be converted
        minify: bool
            If the result should be minified
    """
    if minify:
        def ret(nodes, wrapped): return "".join(nodes)
    else:
        def ret(nodes, wrapped):
            """
                Indent nodes depending on value of wrapped and indent
                If not wrapped, then don't indent
                Otherwise,
                    Seperate each child by a newline
                    and indent each line in the child by one indent unit
            """
            def eachline(nodes):
                """Yield each line in each node"""
                for node in nodes:
                    for line in node.split("\n"):
                        yield line
            if wrapped:
                seperator = "\n{0}".format("    ")
                surrounding = "\n{0}{{0}}\n".format("    ")
            else:
                seperator = "\n"
                surrounding = "{0}"
            return surrounding.format(seperator.join(eachline(nodes)))

    return Node(  # wrap should be app.id
        wrap="nasse", data=data, iterables_repeat_wrap=True
    ).serialize(ret)
